#  Program: parseinput.py
#  Author:  Christopher Hanley
#
#  History:
#   Version 0.1,  11/02/2004: Initial Creation -- CJH
#   Version 0.1.2 01/10/2005: Removed the appending of "_drz.fits" to extracted
#       file names.  -- CJH
#   Version 0.1.3 01/18/2005: Added the NICMOS '_asc.fits' to the list of
#       valid association file names.
#   Version 0.1.4 01/25/2005: Removed reliance on ASN dict keys for ordering
#                   the output filelist. WJH/CJH
#   Version 0.1.5 10/11/2005: Corrected typo in errorstr variable name discovered
#                   by external user j.e.geach@durham.ac.uk.

# irafglob provides the ability to recursively parse user input that
# is in the form of wildcards and '@' files.
from .irafglob import irafglob

from . import fileutil
from stsci.tools.asnutil import readASNTable

__version__ = '0.1.5 (10/11/2005)'
__author__ = 'Christopher Hanley'


def parseinput(inputlist, outputname=None, atfile=None):
    """
    Recursively parse user input based upon the irafglob
    program and construct a list of files that need to be processed.
    This program addresses the following deficiencies of the irafglob program::

       parseinput can extract filenames from association tables

    Returns
    -------
    This program will return a list of input files that will need to
    be processed in addition to the name of any outfiles specified in
    an association table.

    Parameters
    ----------
    inputlist - string
        specification of input files using either wild-cards, @-file or
        comma-separated list of filenames

    outputname - string
        desired name for output product to be created from the input files

    atfile - object
        function to use in interpreting the @-file columns that gets passed to irafglob

    Returns
    -------
    files - list of strings
        names of output files to be processed
    newoutputname - string
        name of output file to be created.

    See Also
    --------
    stsci.tools.irafglob

    """

    # Initalize some variables
    files = [] # list used to store names of input files
    newoutputname = outputname # Outputname returned to calling program.
                               # The value of outputname is only changed
                               # if it had a value of 'None' on input.


    # We can use irafglob to parse the input.  If the input wasn't
    # an association table, it needs to be either a wildcard, '@' file,
    # or comma seperated list.
    files = irafglob(inputlist, atfile=atfile)

    # Now that we have expanded the inputlist into a python list
    # containing the list of input files, it is necessary to examine
    # each of the files to make sure none of them are association tables.
    #
    # If an association table is found, the entries should be read
    # Determine if the input is an association table
    for file in files:
        if (checkASN(file) == True):
            # Create a list to store the files extracted from the
            # association tiable
            assoclist = []

            # The input is an association table
            try:
                # Open the association table
                assocdict = readASNTable(file, None, prodonly=False)
            except:
                errorstr  = "###################################\n"
                errorstr += "#                                 #\n"
                errorstr += "# UNABLE TO READ ASSOCIATION FILE,#\n"
                errorstr +=  str(file)+'\n'
                errorstr += "# DURING FILE PARSING.            #\n"
                errorstr += "#                                 #\n"
                errorstr += "# Please determine if the file is #\n"
                errorstr += "# in the current directory and    #\n"
                errorstr += "# that it has been properly       #\n"
                errorstr += "# formatted.                      #\n"
                errorstr += "#                                 #\n"
                errorstr += "# This error message is being     #\n"
                errorstr += "# generated from within the       #\n"
                errorstr += "# parseinput.py module.           #\n"
                errorstr += "#                                 #\n"
                errorstr += "###################################\n"
                raise ValueError(errorstr)

            # Extract the output name from the association table if None
            # was provided on input.
            if outputname is None:
                newoutputname = assocdict['output']

            # Loop over the association dictionary to extract the input
            # file names.
            for f in assocdict['order']:
                assoclist.append(fileutil.buildRootname(f))

            # Remove the name of the association table from the list of files
            files.remove(file)
            # Append the list of filenames generated from the association table
            # to the master list of input files.
            files.extend(assoclist)

    # Return the list of the input files and the output name if provided in an association.
    return files, newoutputname


def checkASN(filename):
    """
    Determine if the filename provided to the function belongs to
    an association.

    Parameters
    ----------
    filename: string

    Returns
    -------
    validASN  : boolean value

    """
    # Extract the file extn type:
    extnType = filename[filename.rfind('_')+1:filename.rfind('.')]

    # Determine if this extn name is valid for an assocation file
    if isValidAssocExtn(extnType):
        return True
    else:
        return False


def isValidAssocExtn(extname):
    """
    Determine if the extension name given as input could
    represent a valid association file.

    Parameters
    ----------
    extname : string

    Returns
    -------
    isValid  : boolean value

    """
    # Define a list of valid extension types to define an association table.
    validExtnNames = ['asn','asc']

    # Loop over the list of valid extension types and compare with the input
    # extension name.  If there is ever a match return True.
    for validName in validExtnNames:
        if (extname == validName):
            return True
    return False

def countinputs(inputlist):
    """
    Determine the number of inputfiles provided by the user and the
    number of those files that are association tables

    Parameters
    ----------
    inputlist   : string
        the user input

    Returns
    -------
    numInputs: int
        number of inputs provided by the user
    numASNfiles: int
        number of association files provided as input
    """

    # Initialize return values
    numInputs = 0
    numASNfiles = 0

    # User irafglob to count the number of inputfiles
    files = irafglob(inputlist, atfile=None)

    # Use the "len" ufunc to count the number of entries in the list
    numInputs = len(files)

    # Loop over the list and see if any of the entries are association files
    for file in files:
        if (checkASN(file) == True):
            numASNfiles += 1

    return numInputs,numASNfiles



""" Usually copying to and from the clipboard in an app is handled automatically
and correctly on a given platform, when the user applies the right keystrokes
or mouse events for that platform.  In some corner cases this might not be
true, so this module exists to help facilitate any needed copying or pasting.
For now, this is tkinter based, but it is imported lazily.

$Id$
"""
from . import irafutils

_theRoot = None
_lastSel = ''  # our own copy of the last selected text (for PRIMARY)


# Install our own PRIMARY request handler.
def ch_handler(offset=0, length=-1, **kw):
    """ Handle standard PRIMARY clipboard access.  Note that offset and length
    are passed as strings.  This differs from CLIPBOARD. """
    global _lastSel

    offset = int(offset)
    length = int(length)
    if length < 0: length = len(_lastSel)
    return _lastSel[offset:offset+length]


# X11 apps (e.g. xterm) seem to use PRIMARY for select=copy and midmouse=paste
# Other X11 apps        seem to use CLIPBOARD for ctl-c=copy and ?ctl-v?=paste
# OS X seems to use CLIPBOARD for everything, which is Cmd-C and Cmd-V
# Described here:  http://wiki.tcl.tk/1217 "Primary Transfer vs. the Clipboard"
# See also:  http://www.tcl.tk/man/tcl8.5/TkCmd/selection.htm
#      and:  http://www.tcl.tk/man/tcl8.5/TkCmd/clipboard.htm


def put(text, cbname):
    """ Put the given string into the given clipboard. """
    global _lastSel
    _checkTkInit()
    if cbname == 'CLIPBOARD':
        _theRoot.clipboard_clear()
        if text:
            # for clipboard_append, kwds can be -displayof, -format, or -type
            _theRoot.clipboard_append(text)
        return
    if cbname == 'PRIMARY':
        _lastSel = text
        _theRoot.selection_handle(ch_handler, selection='PRIMARY')
        # we need to claim/own it so that ch_handler is used
        _theRoot.selection_own(selection='PRIMARY')
        # could add command arg for a func to be called when we lose ownership
        return
    raise RuntimeError("Unexpected clipboard name: "+str(cbname))


def get(cbname):
    """ Get the contents of the given clipboard. """
    _checkTkInit()
    if cbname == 'PRIMARY':
        try:
            return _theRoot.selection_get(selection='PRIMARY')
        except:
            return None
    if cbname == 'CLIPBOARD':
        try:
            return _theRoot.selection_get(selection='CLIPBOARD')
        except:
            return None
    raise RuntimeError("Unexpected clipboard name: "+str(cbname))


def dump():
    _checkTkInit()
    print ('primary   = '+str(get('PRIMARY')))
    print ('clipboard = '+str(get('CLIPBOARD')))
    print ('owner     = '+str(_theRoot.selection_own_get()))


def _checkTkInit():
    """ Make sure the tkinter root is defined. """
    global _theRoot
    _theRoot = irafutils.init_tk_default_root()



#   Program:    versionInfo.py
#   Author:     Christopher Hanley

#
#   Date:       19 January 2004
#   Purpose:
#       To print a user's system information when providing user support.
#
#   Version:
#       Version 0.1.0, 19-Jan-04: Program created. -- CJH
#       Version 0.1.1, 20-Jan-04: Modified program to
#                                 loop over a taskList object. -- CJH
#
#       Version 0.2.0, 31-Mar-06: Added numpy to the task list. -- CJH


def printVersionInfo():
    # Print the current path information
    try:
        print("Path information:")
        print("-----------------")
        import sys
        print(sys.path)
        print(" ")
    except Exception:
        print("Unable to get sys information.")
        print(" ")

    # Define the list of tasks to test
    taskList = ['numpy',
                'astropy',
                'scipy',
                'matplotlib',
                'drizzlepac',
                'stsci.tools',
                'calcos']

    # Test the list of software tasks
    for software in taskList:
        print(software + ":")
        print("-----------")
        try:
            package = __import__(software)
            try:
                print("version -> ", package.__version__)
            except Exception:
                print("__version__ attribute is not defined")
            try:
                pathName = package.__path__
            except Exception:
                pathName = package.__file__
            print("location -> ", pathName)
        except Exception:
            print(software + " not found in path...")
        print(" ")

    # Print instruction message.
    print("PLEASE PASTE THE OUTPUT FROM THIS TASK ")
    print("AND SEND IT WITH YOUR PROBLEM DESCRIPTION TO")
    print("https://hsthelp.stsci.edu ")


if __name__ == '__main__':
    printVersionInfo()



import multiprocessing
import time
import pytest
from itertools import product

from astropy.io import fits
from astropy.utils.data import get_pkg_data_filename

from stsci.tools.mputil import launch_and_wait, best_tile_layout


SUPPORTED_START_METHODS = []
for sm in ['spawn', 'fork', 'forkserver']:
    try:
        multiprocessing.get_context(sm)
        SUPPORTED_START_METHODS.append(sm)
    except ValueError:
        pass


def takes_time(x, img):
    """Example function which takes some time to run."""
    time.sleep(0.001)  # 1 ms is long by computer standards?


@pytest.mark.parametrize('fname, method', (
    x for x in product([None, 'data/o4sp040b0_raw.fits'],
                       SUPPORTED_START_METHODS)
))
def test_launch_and_wait(fname, method):
    """Illustrate use of launch_and_wait"""
    p = None
    subprocs = []
    # Passing fits.HDUList in Python 3.8 would cause crash due to spawn start
    # method:
    #
    if fname is None:
        img = None
    else:
        img = fits.open(get_pkg_data_filename(fname))

    for item in range(2, 5):
        mp_ctx = multiprocessing.get_context(method)
        p = mp_ctx.Process(target=takes_time, args=(item, img),
                           name='takes_time()')
        subprocs.append(p)

    if method != 'fork' and fname:
        with pytest.raises(TypeError):
            # launch em, pool-fashion
            launch_and_wait(subprocs, 3)
    else:
        launch_and_wait(subprocs, 3)

    if img:
        img.close()


def test_best_tile_layout():
    """Loop though some numbers and make sure we get expected results."""
    for i in range(257):
        x, y = best_tile_layout(i)
        assert (x * y <= i) or (i == 0), \
            "Total num resulting tiles > pool_size"
        unused_cores = i - (x * y)
        # print(i, (x,y), unused_cores)
        if i < 10:
            assert unused_cores <= 1, "Too many idle cores at i = " + str(i)
        else:
            percent_unused = 100. * ((unused_cores * 1.) / i)
            assert percent_unused < 14., "Too many idles cores at i: " + str(i)



#!/usr/bin/env python

# $Id$

"""
Convert a waivered FITS file to various other formats.

Syntax for the command line::

    convertwaiveredfits.py [-hm] [-o <outputFileName>,...] FILE ...

Convert the waivered FITS files (wFITS) to various formats.
The default conversion format is multi-extension FITS (MEF).

Options::

    -h, --help                       print this help message and exit

    -v, --verbose                    turn on verbose output

    -m, --multiExtensionConversion   convert to multi-extension
                                     FITS format (Default)

    -o, --outputFileName             comma separated list of
                                     output file specifications
                                     one per input FILE
                                     Default: input file
                                     specification with the last
                                     character of the base name
                                     changed to `h` in multi-extension FITS format

For example, conversion of a WFPC2 waivered FITS file obtained
from the MAST archive::

    convertwaiveredfits u9zh010bm_c0f.fits

This will convert the waivered FITS file ``u9zh010bm_c0f.fits``
to multi-extension FITS format and generate the output file
``u9zh010bm_c0h.fits``.

Conversion of multiple FITS files can be done using::

    convertwaiveredfits -o out1.fits,out2.fits u9zh010bm_c0f.fits u9zh010bm_c1f.fits

This will convert the waivered FITS files ``u9zh010bm_c0f.fits``
and ``u9zh010bm_c1f.fits`` to multi-extension FITS format and
generate the output files ``out1.fits`` and ``out2.fits``.

Parameters
==========
waiveredObject: obj
    input object representing a waivered FITS
    file; either a ``astropy.io.fits.HDUList`` object, a file
    object, or a file specification.

outputFileName : string
    file specification for the output file.
    Default: `None` - do not generate an output file

forceFileOutput: boolean
    force the generation of an output file when
    the ``outputFileName`` parameter is `None`; the
    output file specification will be the same as
    the input file specification with the last
    character of the base name replaced with the
    character ``h`` in multi-extension FITS format.

    Default: False

convertTo: string
    target conversion type.
    Default: 'multiExtension'

verbose: boolean
    provide verbose output.
    Default: `False`

Returns
=======
hduList : fits.HDUList
    ``astropy.io.fits`` multi-extension FITS object containing converted output

Examples
========

>>> from stsci.tools import convertwaiveredfits
>>> hdulist = convertwaiveredfits.convertwaiveredfits(
...     'u9zh010bm_c0f.fits', forceFileOutput=True)  # doctest: +SKIP

This will convert the waivered FITS file ``u9zh010bm_c0f.fits``
to multi-extension FITS format and write the output to the
file ``u9zh010bm_c0h.fits``;  the returned ``HDUList`` is in
multi-extension FITS format.

>>> from stsci.tools import convertwaiveredfits
>>> with open('u9zh010bm_c0f.fits', mode='rb') as inFile:
...     hdulist = convertwaiveredfits.convertwaiveredfits(inFile, 'out.fits')  # doctest: +SKIP

This will convert the waivered FITS file ``u9zh010bm_c0f.fits``
to multi-extension FITS format and write the output to the
file ``out.fits``; the returned ``HDUList`` is in multi-extension
FITS format.

>>> from astropy.io import fits
>>> from stsci.tools import convertwaiveredfits
>>> with fits.open('u9zh010bm_c0f.fits') as inHdul:
...     hdulist = convertwaiveredfits.convertwaiveredfits(inHdul)  # doctest: +SKIP

This will convert the waivered FITS file ``u9zh010bm_c0f.fits``
to multi-extension FITS format; no output file is generated;
the returned ``HDUList`` is in multi-extension format.

"""
import os
import sys
from astropy.io import fits

__version__ = "1.1 (15 June, 2015)"


#
# -----------------------------------------------------------------------------
# Function definitions
# -----------------------------------------------------------------------------
#
def _usage():
    """
        Print a usage message.

        Parameters: NONE

        Returns: None

        Exceptions: NONE
    """

    print("usage: convertwaiveredfits.py [-hmv] [-o <outputFileName>, ...] FILE ...")


def _processCommandLineArgs():
    """
        Get the command line arguments

        Parameters: NONE

        Returns:

           files            list of file specifications to be converted

           outputFileNames  list of output file specifications
                             (one per input file)
                             Default: a list of None values (one per input file)

           conversionFormat string indicating the conversion format requested
                             Default: "mulitextension"

           verbose          flag indicating if verbose output is desired
                             Default: False

        Exceptions: NONE
    """

    import getopt

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hvmo:",
                                   ["help",
                                    "verbose",
                                    "multiExtensionConversion",
                                    "outputFileName"])
    except getopt.GetoptError as e:
        print(str(e))
        _usage()
        sys.exit(1)

    conversionFormat = ""
    outputFileNames = []
    verbose = False

    for o, a in opts:
        if o in ("-h", "--help"):
            _usage()
            print("       Convert the waivered FITS Files (FILEs) to various formats.")
            print("       The default conversion format is multi-extension FITS.")
            print("       Options:")
            print("         -h,  --help                       display this help message and exit")
            print("         -v,  --verbose                    provide verbose output")
            print("         -m,  --multiExtensionConversion   convert to multiExtension FITS format")
            print("         -o,  --outputFileName             comma separated list of output file")
            print("                                           specifications (one per input FILE)")
            sys.exit()

        if o in ("-v", "--verbose"):
            verbose = True

        if o in ("-m", "--multiExtensionConversion"):
            if conversionFormat != "":
                print("convertwaiveredfits.py: only one conversion format allowed")
                _usage()
                sys.exit(1)

            conversionFormat = "multiExtension"

        if o in ("-o", "--outputFileName"):
            outputFileNames = a.split(',')

    if conversionFormat == "":
        #
        # Set the default conversion format if none was provided
        #
        conversionFormat = "multiExtension"

    if not args:
        print("convertwaiveredfits.py: nothing to convert")
        _usage()
        sys.exit(1)
    else:
        files = args

        if outputFileNames:
            if len(files) != len(outputFileNames):
                print("convertwaiveredfits.py: number of output file names does not match")
                print("                        the number of FILEs to convert")
                _usage()
                sys.exit(1)
        else:
            for i in range(0,len(files)):
                outputFileNames.append(None)

    return files,outputFileNames,conversionFormat,verbose

def _verify(waiveredHdul):
    """
        Verify that the input HDUList is for a waivered FITS file.

        Parameters:

           waiveredHdul     HDUList object to be verified

        Returns: None

        Exceptions:

           ValueError       Input HDUList is not for a waivered FITS file
    """

    if len(waiveredHdul) == 2:
        #
        # There must be exactly 2 HDU's
        #
        if waiveredHdul[0].header['NAXIS'] > 0:
            #
            # The Primary HDU must have some data
            #
            if isinstance(waiveredHdul[1], fits.TableHDU):
                #
                # The Alternate HDU must be a TableHDU
                #
                if waiveredHdul[0].data.shape[0] == \
                   waiveredHdul[1].data.shape[0] or \
                   waiveredHdul[1].data.shape[0] == 1:
                    #
                    # The number of arrays in the Primary HDU must match
                    # the number of rows in the TableHDU.  This includes
                    # the case where there is only a single array and row.
                    #
                    return
    #
    # Not a valid waivered Fits file
    #
    raise ValueError("Input object does not represent a valid waivered" + \
                      " FITS file")

def toMultiExtensionFits(waiveredObject,
                         multiExtensionFileName=None,
                         forceFileOutput=False,
                         verbose=False):
    """
        Convert the input waivered FITS object to a multi-extension FITS
        HDUList object.  Generate an output multi-extension FITS file if
        requested.

        Parameters:

          waiveredObject  input object representing a waivered FITS file;
                          either a astroyp.io.fits.HDUList object, a file object, or a
                          file specification

          outputFileName  file specification for the output file
                          Default: None - do not generate an output file

          forceFileOutput force the generation of an output file when the
                          outputFileName parameter is None; the output file
                          specification will be the same as the input file
                          specification with the last character of the base
                          name replaced with the character 'h'.
                          Default: False

          verbose         provide verbose output
                          Default: False

        Returns:

          mhdul           an HDUList object in multi-extension FITS format.

        Exceptions:

          TypeError       Input object is not a HDUList, a file object or a
                          file name
    """

    if isinstance(waiveredObject, fits.HDUList):
        whdul = waiveredObject
        inputObjectDescription = "HDUList object"
    else:
        try:
            whdul = fits.open(waiveredObject)
            if isinstance(waiveredObject, str):
                inputObjectDescription = "file " + waiveredObject
            else:
                inputObjectDescription = "file " + waiveredObject.name
        except TypeError:
            raise TypeError("Input object must be HDUList, file object, " + \
                            "or file name")

    _verify(whdul)

    undesiredPrimaryHeaderKeywords = ['ORIGIN','FITSDATE','FILENAME',
                                      'ALLG-MAX','ALLG-MIN','ODATTYPE',
                                      'SDASMGNU','OPSIZE','CTYPE2',
                                      'CD2_2','CD2_1','CD1_2','CTYPE3',
                                      'CD3_3','CD3_1','CD1_3','CD2_3',
                                      'CD3_2']
    #
    # Create the multi-extension primary header as a copy of the
    # wavered file primary header
    #
    mPHeader = whdul[0].header
    originalDataType =  whdul[0].header.get('ODATTYPE','')
    #
    # Remove primary header cards with keywords matching the
    # list of undesired primary header keywords
    #
    for keyword in undesiredPrimaryHeaderKeywords:
        #
        # Be careful only to delete the first card that matches
        # the keyword, not all of the cards
        #
        if keyword in mPHeader:
            del mPHeader[mPHeader.index(keyword)]
    #
    # Get the columns from the secondary HDU table
    #
    wcols = whdul[1].columns
    #
    # Remove primary header cards with keywords matching the
    # column names in the secondary HDU table
    #
    for keyword in wcols.names:
        if keyword in mPHeader:
            del mPHeader[keyword]
    #
    # Create the PrimaryHDU
    #
    mPHdu = fits.PrimaryHDU(header=mPHeader)
    #
    # Add the EXTEND card
    #
    mPHdu.header.set('EXTEND', value=True, after='NAXIS')
    #
    # Add the NEXTEND card.  There will be one extension
    # for each row in the wavered Fits file table HDU.
    #
    mPHdu.header['NEXTEND'] = (whdul[1].data.shape[0],
                               'Number of standard extensions')
    #
    # Create the multi-extension file HDUList from the primary header
    #
    mhdul = fits.HDUList([mPHdu])
    #
    # Create the extension HDUs for the multi-extension file.  There
    # will be one extension for each row in the wavered file's table.
    #
    instrument = mPHeader.get('INSTRUME', '')
    nrows = whdul[1].data.shape[0]

    for i in range(0,nrows):
        #
        # Create the basic HDU from the data
        #
        if nrows == 1:
            #
            # Handle case where there is only one row in the table
            #
            data = whdul[0].data
        else:
            data = whdul[0].data[i]

        mhdul.append(fits.ImageHDU(data))
        #
        # Add cards to the header for each keyword in the column
        # names of the secondary HDU table from the wavered file
        #
        for keyword,format,unit in zip(wcols.names,wcols.formats,wcols.units):
            if unit == 'LOGICAL-':
                #
                # Handle logical values
                #
                if whdul[1].data.field(keyword)[i].strip() == 'T':
                    d = True
                else:
                    d = False
            elif format[0] == 'E':
                #
                # Handle floating point values
                #
                fmt = '%'+format[1:]+'G'
                d = eval(fmt % float(whdul[1].data.field(keyword)[i]))  # nosec
            else:
                d = whdul[1].data.field(keyword)[i]

            kw_descr = ""
            if keyword in whdul[1].header:
                kw_descr = whdul[1].header[keyword]
            mhdul[i+1].header[keyword] = (d, kw_descr)
        #
        # If original data is unsigned short then scale the data.
        #
        if originalDataType == 'USHORT':
            mhdul[i+1].scale('int16','',bscale=1,bzero=32768)
            mhdul[i+1].header.set('BSCALE', value=1, before='BZERO')
        #
        # For WFPC2 and FOS instruments require additional header cards
        #
        if instrument in ('WFPC2','FOC'):
            #
            # Add EXTNAME card to header
            #
            mhdul[i+1].header['EXTNAME'] = (mPHeader.get('FILETYPE',''),
                                            'extension name')
            #
            # Add EXTVER card to the header
            #
            mhdul[i+1]._extver = i+1
            mhdul[i+1].header.set('EXTVER', value=i+1,
                                  comment='extension version number',
                                  after='EXTNAME')
            #
            # Add the EXPNAME card to the header
            #
            mhdul[i+1].header.set('EXPNAME',
                                  mPHeader.get('ROOTNAME', ''),
                                  '9 character exposure identifier',
                                  before='EXTVER')
            #
            # Add the INHERIT card to the header.
            #
            mhdul[i+1].header.set('INHERIT', True,
                                  'inherit the primary header',
                                  after='EXTVER')
            #
            # Add the ROOTNAME card to the header
            #
            mhdul[i+1].header.set('ROOTNAME',
                                  mPHeader.get('ROOTNAME', ''),
                                  'rootname of the observationset',
                                  after='INHERIT')

    if not multiExtensionFileName and forceFileOutput:
        base,ext = os.path.splitext(whdul[0]._file.name)
        multiExtensionFileName = base[:-1]+'h'+ext

    verboseString = "Input " + inputObjectDescription + \
                    " converted to multi-extension FITS format."

    if multiExtensionFileName:
        if instrument in ('WFPC2','FOC'):
            #
            # write the FILENAME card to the header for the WFPC2 and FOC
            # instruments
            #
            head,tail = os.path.split(multiExtensionFileName)
            mhdul[0].header.set('FILENAME', value=tail, after='NEXTEND')

        mhdul.writeto(multiExtensionFileName, overwrite=True)
        verboseString = verboseString[:-1] + " and written to " + \
                        multiExtensionFileName + "."

    if verbose:
        print(verboseString)

    return mhdul


def convertwaiveredfits(waiveredObject,
                        outputFileName=None,
                        forceFileOutput=False,
                        convertTo='multiExtension',
                        verbose=False):
    """
        Convert the input waivered FITS object to various formats.  The
        default conversion format is multi-extension FITS.  Generate an output
        file in the desired format if requested.

        Parameters:

          waiveredObject  input object representing a waivered FITS file;
                          either a astropy.io.fits.HDUList object, a file object, or a
                          file specification

          outputFileName  file specification for the output file
                          Default: None - do not generate an output file

          forceFileOutput force the generation of an output file when the
                          outputFileName parameter is None; the output file
                          specification will be the same as the input file
                          specification with the last character of the base
                          name replaced with the character `h` in
                          multi-extension FITS format.

                          Default: False

          convertTo       target conversion type
                          Default: 'multiExtension'

          verbose         provide verbose output
                          Default: False

        Returns:

          hdul            an HDUList object in the requested format.

        Exceptions:

           ValueError       Conversion type is unknown
    """

    if convertTo == 'multiExtension':
        func = toMultiExtensionFits
    else:
        raise ValueError('Conversion type ' + convertTo + ' unknown')

    return func(*(waiveredObject,outputFileName,forceFileOutput,verbose))
#
# *****************************************************************************
# Main Program callable from the shell
# *****************************************************************************
#

def main() :
    files,outputFiles,conversionFormat,verbose = _processCommandLineArgs()

    for f,outputfile in zip(files,outputFiles):
        convertwaiveredfits(f,outputfile,True,conversionFormat,verbose)

    sys.exit()


if __name__ == '__main__':
    main()

"""

Copyright (C) 2005 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""



####
#       Class AlertDialog
#
#       Purpose
#       -------
#
#       AlertDialog's are widgets that allow one to pop up warnings, one line
#       questions etc. They return a set of standard action numbers being :-
#       0 => Cancel was pressed
#       1 => Yes was pressed
#       2 => No was pressed
#
#       Standard Usage
#       --------------
#
#       F = AlertDialog(widget, message)
#       action = F.Show()
####
"""
$Id$
"""
from .dialog import *


class AlertDialog(ModalDialog):

    def __init__(self, widget, msg):
        self.widget = widget
        self.msgString = msg
        Dialog.__init__(self, widget)

    def SetupDialog(self):
        upperFrame = Frame(self.top)
        upperFrame['relief'] = 'raised'
        upperFrame['bd']         = 1
        upperFrame.pack({'expand':'yes', 'side':'top', 'fill':'both' })
        self.bitmap = Label(upperFrame)
        self.bitmap.pack({'side':'left'})
        msgList = self.msgString.split("\n")
        for i in range(len(msgList)):
            msgText = Label(upperFrame)
            msgText["text"]   = msgList[i]
            msgText.pack({'expand':'yes', 'side':'top', 'anchor':'nw',
                    'fill':'x' })
        self.lowerFrame = Frame(self.top)
        self.lowerFrame['relief'] = 'raised'
        self.lowerFrame['bd']    = 1
        self.lowerFrame.pack({'expand':'yes', 'side':'top', 'pady':'2',
                'fill':'both' })

    def OkPressed(self):
        self.TerminateDialog(1)

    def CancelPressed(self):
        self.TerminateDialog(0)

    def NoPressed(self):
        self.TerminateDialog(2)

    def CreateButton(self, text, command):
        self.button = Button(self.lowerFrame)
        self.button["text"]       = text
        self.button["command"]   = command
        self.button.pack({'expand':'yes', 'pady':'2', 'side':'left'})

####
#       Class ErrorDialog
#
#       Purpose
#       -------
#
#       To pop up an error message
####

class ErrorDialog(AlertDialog):

    def SetupDialog(self):
        AlertDialog.SetupDialog(self)
        self.bitmap['bitmap'] = 'error'
        self.CreateButton("OK", self.OkPressed)

####
#       Class WarningDialog
#
#       Purpose
#       -------
#
#       To pop up a warning message.
####

class WarningDialog(AlertDialog):

    def SetupDialog(self):
        AlertDialog.SetupDialog(self)
        self.bitmap['bitmap'] = 'warning'
        self.CreateButton("Yes", self.OkPressed)
        self.CreateButton("No", self.CancelPressed)

####
#       Class QuestionDialog
#
#       Purpose
#       -------
#
#       To pop up a simple question
####

class QuestionDialog(AlertDialog):

    def SetupDialog(self):
        AlertDialog.SetupDialog(self)
        self.bitmap['bitmap'] = 'question'
        self.CreateButton("Yes", self.OkPressed)
        self.CreateButton("No", self.NoPressed)
        self.CreateButton("Cancel", self.CancelPressed)

####
#       Class MessageDialog
#
#       Purpose
#       -------
#
#       To pop up a message.
####

class MessageDialog(AlertDialog):

    def SetupDialog(self):
        AlertDialog.SetupDialog(self)
        self.bitmap['bitmap'] = 'warning'
        self.CreateButton("Dismiss", self.CancelPressed)



from astropy.io import fits

__version__ = '0.3 (01-July-2014)'


class IterFitsFile:
    """ This class defines an object which can be used to
        access the data from a FITS file without leaving
        the file-handle open between reads.

    """
    def __init__(self, name):
        self.name = name
        self.fname = None
        self.extn = None
        self.handle = None
        self.inmemory = False
        self.compress = False

        if not self.fname:
            self.fname,self.extn = parseFilename(name)

    def set_inmemory(self,val):
        """Sets inmemory attribute to either True or False """
        if type(val) is not bool:
            raise TypeError('Input value must be boolean')
        self.inmemory = val

    def _shape(self):
        """ Returns the shape of the data array associated with this file."""
        hdu = self.open()
        _shape = hdu.shape
        if not self.inmemory:
            self.close()
            del hdu
        return _shape

    def _data(self):
        """ Returns the data array associated with this file/extenstion."""
        hdu = self.open()
        _data = hdu.data.copy()
        if not self.inmemory:
            self.close()
            del hdu
        return _data

    def type(self):
        """ Returns the shape of the data array associated with this file."""
        hdu = self.open()
        _type = hdu.data.dtype.name
        if not self.inmemory:
            self.close()
            del hdu
        return _type

    def open(self):
        """ Opens the file for subsequent access. """

        if self.handle is None:
            self.handle = fits.open(self.fname, mode='readonly')

        if self.extn:
            if len(self.extn) == 1:
                hdu = self.handle[self.extn[0]]
            else:
                hdu = self.handle[self.extn[0],self.extn[1]]
        else:
            hdu = self.handle[0]
        if isinstance(hdu,fits.hdu.compressed.CompImageHDU):
            self.compress = True
        return hdu


    def close(self):
        """ Closes file handle for this FITS object."""
        if self.handle is not None:
            self.handle.close()
        self.handle = None

    def __getitem__(self,i):
        """ Returns a PyFITS section for the rows specified. """
        # All I/O must be done here, starting with open
        hdu = self.open()
        if self.inmemory or self.compress:
            _data = hdu.data[i,:]
        else:
            _data = hdu.section[i,:]

        if not self.inmemory:
            self.close()
            del hdu

        return _data


    def __getattribute__(self,name):
        if name == 'data':
            return self._data()
        elif name == 'shape':
            return self._shape()
        else:
            return object.__getattribute__(self,name)


def parseFilename(filename):
    """
        Parse out filename from any specified extensions.
        Returns rootname and string version of extension name.

        Modified from 'pydrizzle.fileutil' to allow this
        module to be independent of PyDrizzle/MultiDrizzle.

    """
    # Parse out any extension specified in filename
    _indx = filename.find('[')
    if _indx > 0:
        # Read extension name provided
        _fname = filename[:_indx]
        extn = filename[_indx+1:-1]

        # An extension was provided, so parse it out...
        if repr(extn).find(',') > 1:
            _extns = extn.split(',')
            # Two values given for extension:
            #    for example, 'sci,1' or 'dq,1'
            _extn = [_extns[0],int(_extns[1])]
        elif repr(extn).find('/') > 1:
            # We are working with GEIS group syntax
            _indx = str(extn[:extn.find('/')])
            _extn = [int(_indx)]
        elif isinstance(extn, str):
            # Only one extension value specified...
            if extn.isdigit():
                # We only have an extension number specified as a string...
                _nextn = int(extn)
            else:
                # We only have EXTNAME specified...
                _nextn = extn
            _extn = [_nextn]
        else:
            # Only integer extension number given, or default of 0 is used.
            _extn = [int(extn)]

    else:
        _fname = filename
        _extn = None
    return _fname,_extn



from stsci.tools import parseinput, fileutil, convertwaiveredfits, readgeis
from astropy.io import fits
import os


def checkFiles(filelist,ivmlist = None):
    """
    - Converts waiver fits sciece and data quality files to MEF format
    - Converts GEIS science and data quality files to MEF format
    - Checks for stis association tables and splits them into single imsets
    - Removes files with EXPTIME=0 and the corresponding ivm files
    - Removes files with NGOODPIX == 0 (to exclude saturated images)
    - Removes files with missing PA_V3 keyword

    The list of science files should match the list of ivm files at the end.
    """
    toclose = False
    if isinstance(filelist[0], str):
        toclose = True
    newfilelist, ivmlist = checkFITSFormat(filelist, ivmlist)

    # check for STIS association files. This must be done before
    # the other checks in order to handle correctly stis
    # assoc files
    newfilelist, ivmlist = checkStisFiles(newfilelist, ivmlist)
    if newfilelist == []:
        return [], []
    removed_expt_files = check_exptime(newfilelist)

    newfilelist, ivmlist = update_input(newfilelist, ivmlist, removed_expt_files)
    if newfilelist == []:
        return [], []
    removed_ngood_files = checkNGOODPIX(newfilelist)
    newfilelist, ivmlist = update_input(newfilelist, ivmlist, removed_ngood_files)
    if newfilelist == []:
        return [], []

    removed_pav3_files = checkPA_V3(newfilelist)
    newfilelist, ivmlist = update_input(newfilelist, ivmlist, removed_pav3_files)

    newfilelist, ivmlist = update_input(newfilelist, ivmlist,[])

    if newfilelist == []:
        return [], []

    if toclose:
        newfilelist = [hdul.filename() for hdul in newfilelist]
    return newfilelist, ivmlist

def checkFITSFormat(filelist, ivmlist=None):
    """
    This code will check whether or not files are GEIS or WAIVER FITS and
    convert them to MEF if found. It also keeps the IVMLIST consistent with
    the input filelist, in the case that some inputs get dropped during
    the check/conversion.
    """
    if ivmlist is None:
        ivmlist = [None for l in filelist]

    sci_ivm = list(zip(filelist, ivmlist))

    removed_files, translated_names, newivmlist = convert2fits(sci_ivm)
    newfilelist, ivmlist = update_input(filelist, ivmlist, removed_files)

    if newfilelist == [] and translated_names == []:
        return [], []

    elif translated_names != []:
        newfilelist.extend(translated_names)
        ivmlist.extend(newivmlist)

    return newfilelist, ivmlist


def checkStisFiles(filelist, ivmlist=None):
    newflist = []
    newilist = []
    removed_files = []
    assoc_files = []
    assoc_ilist = []

    if len(filelist) != len(ivmlist):
        errormsg = "Input file list and ivm list have different lenghts\n"
        errormsg += "Quitting ...\n"
        raise ValueError(errormsg)

    toclose = False
    for t in zip(filelist, ivmlist):
        if isinstance(t[0], str):
            t = (fits.open(t[0]), t[1])
            toclose = True
        if t[0][0].header['INSTRUME'] != 'STIS':
            newflist.append(t[0])
            newilist.append(t[1])
            continue
        if isSTISSpectroscopic(t[0]):
            removed_files.append(t[0])
            continue
        sci_count = stisObsCount(t[0])
        stisExt2PrimKw([t[0]])
        if sci_count >1:
            newfilenames = splitStis(t[0], sci_count)

            assoc_files.extend(newfilenames)
            removed_files.append(t[0])
            if (isinstance(t[1], tuple) and t[1][0] is not None) or \
               (not isinstance(t[1], tuple) and t[1] is not None):
                print('Does not handle STIS IVM files and STIS association files\n')
            else:
                asn_ivmlist = list(zip(sci_count * [None], newfilenames))
                assoc_ilist.extend(asn_ivmlist)
        elif sci_count == 1:
            newflist.append(t[0])
            newilist.append(t[1])
        else:
            errormsg = "No valid 'SCI extension in STIS file\n"
            raise ValueError(errormsg)

        if toclose:
            t[0].close()
    newflist.extend(assoc_files)
    newilist.extend(assoc_ilist)
    return newflist, newilist

def check_exptime(filelist):
    """
    Removes files with EXPTIME==0 from filelist.
    """
    toclose = False
    removed_files = []
    for f in filelist:
        if isinstance(f, str):
            f = fits.open(f)
            toclose = True

        try:
            exptime = f[0].header['EXPTIME']
        except KeyError:
            removed_files.append(f)
            print("Warning:  There are files without keyword EXPTIME")
            continue
        if exptime <= 0:
            removed_files.append(f)
            print("Warning:  There are files with zero exposure time: keyword EXPTIME = 0.0")

    if removed_files != []:
        print("Warning:  Removing the following files from input list")
        for f in removed_files:
            print('\t',f.filename() or "")
    return removed_files

def checkNGOODPIX(filelist):
    """
    Only for ACS, WFC3 and STIS, check NGOODPIX
    If all pixels are 'bad' on all chips, exclude this image
    from further processing.
    Similar checks requiring comparing 'driz_sep_bits' against
    WFPC2 c1f.fits arrays and NICMOS DQ arrays will need to be
    done separately (and later).
    """
    toclose = False
    removed_files = []
    supported_instruments = ['ACS','STIS','WFC3']
    for inputfile in filelist:
        if isinstance(inputfile, str):
            if fileutil.getKeyword(inputfile,'instrume') in supported_instruments:
                inputfile = fits.open(inputfile)
                toclose = True
        elif inputfile[0].header['instrume'] not in supported_instruments:
            continue

        ngood = 0
        for extn in inputfile:
            if 'EXTNAME' in extn.header and extn.header['EXTNAME'] == 'SCI':
                ngood += extn.header['NGOODPIX']
        if (ngood == 0):
            removed_files.append(inputfile)
        if toclose:
            inputfile.close()

    if removed_files != []:
        print("Warning:  Files without valid pixels detected: keyword NGOODPIX = 0.0")
        print("Warning:  Removing the following files from input list")
        for f in removed_files:
            print('\t',f.filename() or "")

    return removed_files

def update_input(filelist, ivmlist=None, removed_files=None):
    """
    Removes files flagged to be removed from the input filelist.
    Removes the corresponding ivm files if present.
    """
    newfilelist = []

    if removed_files == []:
        return filelist, ivmlist
    else:
        sci_ivm = list(zip(filelist, ivmlist))
        for f in removed_files:
            result = [sci_ivm.remove(t) for t in sci_ivm if t[0] == f ]
        ivmlist = [el[1] for el in sci_ivm]
        newfilelist = [el[0] for el in sci_ivm]
        return newfilelist, ivmlist


def stisObsCount(input):
    """
    Input: A stis multiextension file
    Output: Number of stis science extensions in input
    """
    count = 0
    toclose = False
    if isinstance(input, str):
        input = fits.open(input)
        toclose = True
    for ext in input:
        if 'extname' in ext.header:
            if (ext.header['extname'].upper() == 'SCI'):
                count += 1
    if toclose:
        input.close()
    return count

def splitStis(stisfile, sci_count):
    """
    Split a STIS association file into multiple imset MEF files.

    Split the corresponding spt file if present into single spt files.
    If an spt file can't be split or is missing a Warning is printed.

    Returns
    -------
    names: list
        a list with the names of the new flt files.

    """
    newfiles = []

    toclose = False
    if isinstance(stisfile, str):
        f = fits.open(stisfile)
        toclose = True
    else:
        f = stisfile
    hdu0 = f[0].copy()
    stisfilename = stisfile.filename()

    for count in range(1,sci_count+1):
        fitsobj = fits.HDUList()
        fitsobj.append(hdu0)
        hdu = f[('sci',count)].copy()
        fitsobj.append(hdu)
        rootname = hdu.header['EXPNAME']
        newfilename = fileutil.buildNewRootname(rootname, extn='_flt.fits')
        try:
            # Verify error array exists
            if f[('err', count)].data is None:
                raise ValueError
            # Verify dq array exists
            if f[('dq', count)].data is None:
                raise ValueError
            # Copy the err extension
            hdu = f[('err',count)].copy()
            fitsobj.append(hdu)
            # Copy the dq extension
            hdu = f[('dq',count)].copy()
            fitsobj.append(hdu)
            fitsobj[1].header['EXTVER'] = 1
            fitsobj[2].header['EXTVER'] = 1
            fitsobj[3].header['EXTVER'] = 1

        except ValueError:
            print('\nWarning:')
            print('Extension version %d of the input file %s does not' %(count, stisfile))
            print('contain all required image extensions. Each must contain')
            print('populates SCI, ERR and DQ arrays.')

            continue


        # Determine if the file you wish to create already exists on the disk.
        # If the file does exist, replace it.
        if (os.path.exists(newfilename)):
            os.remove(newfilename)
            print("       Replacing "+newfilename+"...")

            # Write out the new file
        fitsobj.writeto(newfilename)
        # Insure returned HDUList is associated with a file
        fitsobj.close()
        fitsobj = fits.open(newfilename, mode='update')
        newfiles.append(fitsobj) # Return HDUList, not filename

    f.close()

    sptfilename = fileutil.buildNewRootname(stisfilename, extn='_spt.fits')
    try:
        sptfile = fits.open(sptfilename)
    except IOError:
        print('SPT file not found %s \n' % sptfilename)
        return newfiles

    if sptfile:
        hdu0 = sptfile[0].copy()
        try:
            for count in range(1,sci_count+1):
                fitsobj = fits.HDUList()
                fitsobj.append(hdu0)
                hdu = sptfile[count].copy()
                fitsobj.append(hdu)
                rootname = hdu.header['EXPNAME']
                newfilename = fileutil.buildNewRootname(rootname, extn='_spt.fits')
                fitsobj[1].header['EXTVER'] = 1
                if (os.path.exists(newfilename)):
                    os.remove(newfilename)
                    print("       Replacing "+newfilename+"...")

                # Write out the new file
                fitsobj.writeto(newfilename)
        except:
            print("Warning: Unable to split spt file %s " % sptfilename)
        if toclose:
            sptfile.close()

    return newfiles

def stisExt2PrimKw(stisfiles):
    """
        Several kw which are usually in the primary header
        are in the extension header for STIS. They are copied to
        the primary header for convenience.
        List if kw:
        'DATE-OBS', 'EXPEND', 'EXPSTART', 'EXPTIME'
    """

    kw_list = ['DATE-OBS', 'EXPEND', 'EXPSTART', 'EXPTIME']

    for sfile in stisfiles:
        toclose = False

        if isinstance(sfile, str):
            sfile = fits.open(sfile, mode='update')
            toclose = True

        #d = {}
        for k in kw_list:
            sfile[0].header[k] = sfile[1].header[k]
            sfile[0].header.comments[k] = "Copied from extension header"
        if toclose:
            sfile.close()


def isSTISSpectroscopic(fname):

    if (isinstance(fname, fits.HDUList) and fname[0].header['OBSTYPE'] == 'SPECTROSCOPIC'
        or isinstance(fname, str) and fits.getval(fname, 'OBSTYPE') == 'SPECTROSCOPIC'):
        print("Warning:  STIS spectroscopic files detected")
        print("Warning:  Removing %s from input list" % fname)
        return True
    else:
        return False

def checkPA_V3(fnames):
    removed_files = []
    for f in fnames:
        toclose = False
        if isinstance(f, str):
            f = fits.open(f)
            toclose = True
        try:
            pav3 = f[0].header['PA_V3']
        except KeyError:
            rootname = f[0].header['ROOTNAME']
            sptfile = rootname+'_spt.fits'
            if fileutil.findFile(sptfile):
                try:
                    pav3 = fits.getval(sptfile, 'PA_V3')
                except KeyError:
                    print("Warning:  Files without keyword PA_V3 detected")
                    removed_files.append(f.filename() or "")
                f[0].header['PA_V3'] = pav3
            else:
                print("Warning:  Files without keyword PA_V3 detected")
                removed_files.append(f.filename() or "")
        if toclose:
            f.close()
    if removed_files != []:
        print("Warning:  Removing the following files from input list")
        for f in removed_files:
            print('\t',f)
    return removed_files

def convert2fits(sci_ivm):
    """
    Checks if a file is in WAIVER of GEIS format and converts it to MEF
    """
    removed_files = []
    translated_names = []
    newivmlist = []

    for file in sci_ivm:
        #find out what the input is
        # if science file is not found on disk, add it to removed_files for removal
        try:
            imgfits,imgtype = fileutil.isFits(file[0])
        except IOError:
            print("Warning:  File %s could not be found" %file[0])
            print("Warning:  Removing file %s from input list" %file[0])
            removed_files.append(file[0])
            continue

        # Check for existence of waiver FITS input, and quit if found.
        # Or should we print a warning and continue but not use that file
        if imgfits and imgtype == 'waiver':
            newfilename = waiver2mef(file[0], convert_dq=True)
            if newfilename is None:
                print("Removing file %s from input list - could not convert WAIVER format to MEF\n" %file[0])
                removed_files.append(file[0])
            else:
                removed_files.append(file[0])
                translated_names.append(newfilename)
                newivmlist.append(file[1])

        # If a GEIS image is provided as input, create a new MEF file with
        # a name generated using 'buildFITSName()'
        # Convert the corresponding data quality file if present
        if not imgfits:
            newfilename = geis2mef(file[0], convert_dq=True)
            if newfilename is None:
                print("Removing file %s from input list - could not convert GEIS format to MEF\n" %file[0])
                removed_files.append(file[0])
            else:
                removed_files.append(file[0])
                translated_names.append(newfilename)
                newivmlist.append(file[1])

    return removed_files, translated_names, newivmlist

def waiver2mef(sciname, newname=None, convert_dq=True, writefits=True):
    """
    Converts a GEIS science file and its corresponding
    data quality file (if present) to MEF format
    Writes out both files to disk.
    Returns the new name of the science image.
    """

    if isinstance(sciname, fits.HDUList):
        filename = sciname.filename()
    else:
        filename = sciname

    try:
        clobber = True
        fimg = convertwaiveredfits.convertwaiveredfits(filename)

        #check for the existence of a data quality file
        _dqname = fileutil.buildNewRootname(filename, extn='_c1f.fits')
        dqexists = os.path.exists(_dqname)
        if convert_dq and dqexists:
            try:
                dqfile = convertwaiveredfits.convertwaiveredfits(_dqname)
                dqfitsname = fileutil.buildNewRootname(_dqname, extn='_c1h.fits')
            except Exception:
                print("Could not read data quality file %s" % _dqname)
        if writefits:
            # User wants to make a FITS copy and update it
            # using the filename they have provided
            rname = fileutil.buildNewRootname(filename)
            fitsname = fileutil.buildNewRootname(rname, extn='_c0h.fits')

            # Write out GEIS image as multi-extension FITS.
            fexists = os.path.exists(fitsname)
            if (fexists and clobber) or not fexists:
                print('Writing out WAIVERED as MEF to ', fitsname)
                fimg.writeto(fitsname, overwrite=clobber)
                if dqexists:
                    print('Writing out WAIVERED as MEF to ', dqfitsname)
                    dqfile.writeto(dqfitsname, overwrite=clobber)

        # Now close input GEIS image, and open writable
        # handle to output FITS image instead...
        fimg.close()
        del fimg

        fimg = fits.open(fitsname, mode='update', memmap=False)

        return fimg
    except IOError:
        print('Warning: File %s could not be found' % sciname)
        return None


def geis2mef(sciname, convert_dq=True):
    """
    Converts a GEIS science file and its corresponding
    data quality file (if present) to MEF format
    Writes out both files to disk.
    Returns the new name of the science image.
    """
    clobber = True
    mode = 'update'
    memmap = True
    # Input was specified as a GEIS image, but no FITS copy
    # exists.  Read it in with 'readgeis' and make a copy
    # then open the FITS copy...
    try:
        # Open as a GEIS image for reading only
        fimg = readgeis.readgeis(sciname)
    except Exception:
        raise IOError("Could not open GEIS input: %s" % sciname)

    #check for the existence of a data quality file
    _dqname = fileutil.buildNewRootname(sciname, extn='.c1h')
    dqexists = os.path.exists(_dqname)
    if dqexists:
        try:
            dqfile = readgeis.readgeis(_dqname)
            dqfitsname = fileutil.buildFITSName(_dqname)
        except Exception:
            print("Could not read data quality file %s" % _dqname)

    # Check to see if user wanted to update GEIS header.
    # or write out a multi-extension FITS file and return a handle to it
    # User wants to make a FITS copy and update it
    # using the filename they have provided
    fitsname = fileutil.buildFITSName(sciname)

    # Write out GEIS image as multi-extension FITS.
    fexists = os.path.exists(fitsname)
    if (fexists and clobber) or not fexists:
            print('Writing out GEIS as MEF to ', fitsname)
            fimg.writeto(fitsname, overwrite=clobber)
            if dqexists:
                print('Writing out GEIS as MEF to ', dqfitsname)
                dqfile.writeto(dqfitsname, overwrite=clobber)

    # Now close input GEIS image, and open writable
    # handle to output FITS image instead...
    fimg.close()
    del fimg
    fimg = fits.open(fitsname, mode=mode, memmap=memmap)

    return fimg


def countInput(input):
    files = parseinput.parseinput(input)
    count = len(files[0])
    for f in files[0]:
        if fileutil.isFits(f)[0]:
            try:
                ins = fits.getval(f, 'INSTRUME')
            except Exception:  # allow odd fits files; do not stop the count
                ins = None
            if ins == 'STIS':
                count += (stisObsCount(f)-1)
    return count



"""
A collection of utilities for handling output to standard out/err as well as
to file-based or other logging handlers through a single interface.
"""
import builtins
import inspect
import logging
import os
import sys
import threading

global_logging_started = False


# The global_logging system replaces the input builtin
# because:
#
#  1) It's the easiest way to capture the raw_input prompt and subsequent user
#     input to the log.
def global_logging_raw_input(prompt):
    retval = builtins._original_raw_input(prompt)
    if isinstance(sys.stdout, StreamTeeLogger):
        sys.stdout.log_orig(str(prompt) + retval, echo=False)
    return retval


class StreamTeeLogger(logging.Logger):
    """
    A Logger implementation that is meant to replace an I/O stream such as
    `sys.stdout`, `sys.stderr`, or any other stream-like object that supports a
    `write()` method and a `flush()` method.

    When `StreamTeeLogger.write` is called, the written strings are
    line-buffered, and each line is logged through the normal Python logging
    interface.  The `StreamTeeLogger` has two handlers:

     * The LogTeeHandler redirects all log messages to a logger with the same
       name as the module in which the `write()` method was called.  For
       example, if this logger is used to replace `sys.stdout`, all `print`
       statements in the module `foo.bar` will be logged to a logger called
       ``foo.bar``.

    * If the ``stream`` argument was provided, this logger also attaches a
      `logging.StreamHandler` to itself for the given ``stream``.  For example,
      if ``stream=sys.stdout`` then messages sent to this logger will be output
      to `sys.stdout`.  However, only messages created through the `write()`
      method call will re-output to the given stream.

    Parameters
    ----------
    name : string
        The name of this logger, as in `logging.Logger`

    level : int (optional)
        The minimum level at which to log messages sent to this logger; also
        used as the default level for messages logged through the `write()`
        interface (default: `logging.INFO`).

    stream : stream-like object (optional)
        The stream-like object (an object with `write()` and `flush()`
        methods) to tee to; should be the same file object being replaced (i.e.
        sys.stdout).  If `None` (the default) writes to this file will not be
        sent to a stream logger.

    See Also
    --------
    `EchoFilter` is a logger filter that can control which modules' output is
    sent to the screen via the `StreamHandler` on this logger.
    """

    def __init__(self, name, level=logging.INFO, stream=None):
        logging.Logger.__init__(self, name, level)
        self.__thread_local_ctx = threading.local()
        self.__thread_local_ctx.write_count = 0
        self.propagate = False
        self.buffer = ''

        self.stream = None
        self.set_stream(stream)

        self.addHandler(_LogTeeHandler())
        #self.errors = 'strict'
        #self.encoding = 'utf8'

    @property
    def encoding(self):
        if self.stream:
            try:
                return self.stream.encoding
            except AttributeError:
                pass

        # Default value
        return 'utf-8'

    @property
    def errors(self):
        if self.stream:
            try:
                return self.stream.errors
            except AttributeError:
                pass

        # Default value
        return 'strict'

    def set_stream(self, stream):
        """
        Set the stream that this logger is meant to replace.  Usually this will
        be either `sys.stdout` or `sys.stderr`, but can be any object with
        `write()` and `flush()` methods, as supported by
        `logging.StreamHandler`.
        """

        for handler in self.handlers[:]:
            if isinstance(handler, logging.StreamHandler):
                self.handlers.remove(handler)

        if stream is not None:
            stream_handler = logging.StreamHandler(stream)
            stream_handler.addFilter(_StreamHandlerEchoFilter())
            stream_handler.setFormatter(logging.Formatter('%(message)s'))
            self.addHandler(stream_handler)

        self.stream = stream

    def write(self, message):
        """
        Buffers each message until a newline is reached.  Each complete line is
        then published to the logging system through ``self.log()``.
        """

        self.__thread_local_ctx.write_count += 1

        try:
            if self.__thread_local_ctx.write_count > 1:
                return

            # For each line in the buffer ending with \n, output that line to
            # the logger
            msgs = (self.buffer + message).split('\n')
            self.buffer = msgs.pop(-1)
            for m in msgs:
                self.log_orig(m, echo=True)
        finally:
            self.__thread_local_ctx.write_count -= 1

    def flush(self):
        """
        Flushes all handlers attached to this logger; this includes flushing
        any attached stream-like object (e.g. `sys.stdout`).
        """

        for handler in self.handlers:
            handler.flush()

    def fileno(self):
        fd = None
        if self.stream:
            try:
                fd = self.stream.fileno()
            except:
                fd = None
        if fd is None:
            raise IOError('fileno() not defined for logger stream %r' %
                          self.stream)
        return fd

    def log_orig(self, message, echo=True):
        modname, path, lno, func = self.find_actual_caller()
        self.log(self.level, message,
                 extra={'orig_name': modname, 'orig_pathname': path,
                        'orig_lineno': lno, 'orig_func': func, 'echo': echo})

    def find_actual_caller(self):
        """
        Returns the full-qualified module name, full pathname, line number, and
        function in which `StreamTeeLogger.write()` was called.  For example,
        if this instance is used to replace `sys.stdout`, this will return the
        location of any print statement.
        """

        # Gleaned from code in the logging module itself...
        try:
            f = sys._getframe(1)
            ##f = inspect.currentframe(1)
        except Exception:
            f = None
        # On some versions of IronPython, currentframe() returns None if
         # IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back
        rv = "(unknown module)", "(unknown file)", 0, "(unknown function)"
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            mod = inspect.getmodule(f)

            if mod is None:
                modname = '__main__'
            else:
                modname = mod.__name__

            if modname == __name__:
                # Crawl back until the first frame outside of this module
                f = f.f_back
                continue

            rv = (modname, filename, f.f_lineno, co.co_name)
            break
        return rv


class EchoFilter:
    """
    A logger filter primarily for use with `StreamTeeLogger`.  Adding an
    `EchoFilter` to a `StreamTeeLogger` instances allows control over which
    modules' print statements, for example, are output to stdout.

    For example, to allow only output from the 'foo' module to be printed to
    the console:

    >>> stdout_logger = logging.getLogger('stsci.tools.logutil.stdout')
    >>> stdout_logger.addFilter(EchoFilter(include=['foo']))

    Now only print statements in the 'foo' module (or any sub-modules if 'foo'
    is a package) are printed to stdout.   Any other print statements are just
    sent to the appropriate logger.

    Parameters
    ----------
    include : iterable
        Packages or modules to include in stream output.  If set, then only the
        modules listed here are output to the stream.

    exclude : iterable
        Packages or modules to be excluded from stream output.  If set then all
        modules except for those listed here are output to the stream.  If both
        ``include`` and ``exclude`` are provided, ``include`` takes precedence
        and ``exclude`` is ignored.
    """

    def __init__(self, include=None, exclude=None):
        self.include = set(include) if include is not None else include
        self.exclude = set(exclude) if exclude is not None else exclude

    def filter(self, record):
        if ((self.include is None and self.exclude is None) or
            not hasattr(record, 'orig_name')):
            return True

        record_name = record.orig_name.split('.')
        while record_name:
            if self.include is not None:
                if '.'.join(record_name) in self.include:
                    return True
            elif self.exclude is not None:
                if '.'.join(record_name) not in self.exclude:
                    return True
                else:
                    break
            record_name.pop()

        record.echo = False
        return True


class LoggingExceptionHook:
    def __init__(self, logger, level=logging.ERROR):
        self._oldexcepthook = sys.excepthook
        self.logger = logger
        self.level = level
        if not self.logger.handlers:
            self.logger.addHandler(logging.NullHandler())

    def __del__(self):
        try:
            try:
                sys.excepthook = self._oldexcepthook
            except AttributeError:
                sys.excepthook = sys.__excepthook__
        except AttributeError:
            pass

    def __call__(self, exc_type, exc_value, traceback):
        self.logger.log(self.level, 'An unhandled exception ocurred:',
                        exc_info=(exc_type, exc_value, traceback))
        self._oldexcepthook(exc_type, exc_value, traceback)


def setup_global_logging():
    """
    Initializes capture of stdout/stderr, Python warnings, and exceptions;
    redirecting them to the loggers for the modules from which they originated.
    """

    global global_logging_started

    if global_logging_started:
        return

    orig_logger_class = logging.getLoggerClass()
    logging.setLoggerClass(StreamTeeLogger)
    try:
        stdout_logger = logging.getLogger(__name__ + '.stdout')
        stderr_logger = logging.getLogger(__name__ + '.stderr')
    finally:
        logging.setLoggerClass(orig_logger_class)

    stdout_logger.setLevel(logging.INFO)
    stderr_logger.setLevel(logging.ERROR)
    stdout_logger.set_stream(sys.stdout)
    stderr_logger.set_stream(sys.stderr)
    sys.stdout = stdout_logger
    sys.stderr = stderr_logger

    exception_logger = logging.getLogger(__name__ + '.exc')
    sys.excepthook = LoggingExceptionHook(exception_logger)

    logging.captureWarnings(True)

    rawinput = 'input'
    builtins._original_raw_input = getattr(builtins, rawinput)
    setattr(builtins, rawinput, global_logging_raw_input)

    global_logging_started = True


def teardown_global_logging():
    """Disable global logging of stdio, warnings, and exceptions."""

    global global_logging_started
    if not global_logging_started:
        return

    stdout_logger = logging.getLogger(__name__ + '.stdout')
    stderr_logger = logging.getLogger(__name__ + '.stderr')
    if sys.stdout is stdout_logger:
        sys.stdout = sys.stdout.stream
    if sys.stderr is stderr_logger:
        sys.stderr = sys.stderr.stream

    # If we still have an unhandled exception go ahead and handle it with the
    # replacement excepthook before deleting it
    exc_type, exc_value, exc_traceback = sys.exc_info()
    if exc_type is not None:
        sys.excepthook(exc_type, exc_value, exc_traceback)
    del exc_type
    del exc_value
    del exc_traceback

    del sys.excepthook
    logging.captureWarnings(False)

    rawinput = 'input'
    if hasattr(builtins, '_original_raw_input'):
        setattr(builtins, rawinput, builtins._original_raw_input)
        del builtins._original_raw_input

    global_logging_started = False


# Cribbed, with a few tweaks from Tom Aldcroft at
# http://www.astropython.org/snippet/2010/2/Easier-python-logging
def create_logger(name, format='%(levelname)s: %(message)s', datefmt=None,
                  stream=None, level=logging.INFO, filename=None, filemode='w',
                  filelevel=None, propagate=True):
    """
    Do basic configuration for the logging system. Similar to
    logging.basicConfig but the logger ``name`` is configurable and both a file
    output and a stream output can be created. Returns a logger object.

    The default behaviour is to create a logger called ``name`` with a null
    handled, and to use the "%(levelname)s: %(message)s" format string, and add
    the handler to the ``name`` logger.

    A number of optional keyword arguments may be specified, which can alter
    the default behaviour.

    :param name: Logger name
    :param format: handler format string
    :param datefmt: handler date/time format specifier
    :param stream: add a StreamHandler using ``stream``
                    (None disables the stream, default=None)
    :param level: logger level (default=INFO).
    :param filename: add a FileHandler using ``filename`` (default=None)
    :param filemode: open ``filename`` with specified filemode ('w' or 'a')
    :param filelevel: logger level for file logger (default=``level``)
    :param propagate: propagate message to parent (default=True)

    :returns: logging.Logger object
    """

    # Get a logger for the specified name
    logger = logging.getLogger(name)
    logger.setLevel(level)
    fmt = logging.Formatter(format, datefmt)
    logger.propagate = propagate

    # Remove existing handlers, otherwise multiple handlers can accrue
    for hdlr in logger.handlers:
        logger.removeHandler(hdlr)

    # Add handlers. Add NullHandler if no file or stream output so that
    # modules don't emit a warning about no handler.
    if not (filename or stream):
        logger.addHandler(logging.NullHandler())

    if filename:
        hdlr = logging.FileHandler(filename, filemode)
        if filelevel is None:
            filelevel = level
        hdlr.setLevel(filelevel)
        hdlr.setFormatter(fmt)
        logger.addHandler(hdlr)

    if stream:
        hdlr = logging.StreamHandler(stream)
        hdlr.setLevel(level)
        hdlr.setFormatter(fmt)
        logger.addHandler(hdlr)

    return logger


class _StreamHandlerEchoFilter(logging.Filter):
    """
    Filter used by the `logging.StreamHandler` internal to `StreamTeeLogger`;
    any message logged through `StreamTeeLogger.write()` has an ``echo=True``
    attribute attached to the `LogRecord`.  This ensures that the
    `StreamHandler` only logs messages with this ``echo`` attribute set to
    `True`.
    """

    def filter(self, record):
        if hasattr(record, 'echo'):
            return record.echo
        return False


class _LogTeeHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        logging.Handler.__init__(self, level)
        self.__thread_local_ctx = threading.local()
        self.__thread_local_ctx.logger_handle_counts = {}

    def emit(self, record):
        # Hand off to the global logger with the name same as the module of
        # origin for this record
        if not hasattr(record, 'orig_name'):
            return

        record = logging.LogRecord(record.orig_name, record.levelno,
                                   record.orig_pathname, record.orig_lineno,
                                   record.msg, record.args, record.exc_info,
                                   record.orig_func)
        record.origin = ""
        logger = logging.getLogger(record.name)
        if not logger.handlers:
            logger.addHandler(logging.NullHandler())

        counts = self.__thread_local_ctx.logger_handle_counts
        if logger.name in counts:
            counts[logger.name] += 1
        else:
            counts[logger.name] = 1
            if self._search_stack():
                return
        try:
            if counts[logger.name] > 1:
                return
            logger.handle(record)
        finally:
            counts[logger.name] -= 1

    def _search_stack(self):
        curr_frame = sys._getframe(3)
        ##curr_frame = inspect.currentframe(3)
        while curr_frame:
            if 'self' in curr_frame.f_locals:
                s = curr_frame.f_locals['self']
                if (isinstance(s, logging.Logger) and not
                    isinstance(s, StreamTeeLogger)):
                    return True
            curr_frame = curr_frame.f_back
        return False



"""
Return the gaussian fit of a 1D array.

Uses mpfit.py - a python implementation of the Levenberg-Marquardt
least-squares minimization, based on MINPACK-1. See nmpfit.py for
the history of this module (fortran -> idl -> python).

@author: Nadia Dencheva
@version: '1.0 (2007-02-20)'

"""
import numpy as np

from . import nmpfit

import warnings

warnings.warn("GFIT is deprecated - stsci.tools v 3.4.12 is the last version to contain it."
              "Use astropy.modeling instead.")

__version__ = '2.0'          # Release version number only
__vdate__ = '2018-04-20'     # Date of this version


def _gauss_funct(p, fjac=None, x=None, y=None, err=None,
                 weights=None):

    """
    Defines the gaussian function to be used as the model.

    """
    if p[2] != 0.0:
        Z = (x - p[1]) / p[2]
        model = p[0] * np.e ** (-Z ** 2 / 2.0)
    else:
        model = np.zeros(np.size(x))

    status = 0
    if weights is not None:
        if err is not None:
            print("Warning: Ignoring errors and using weights.\n")

        return [status, (y - model) * weights]

    elif err is not None:
        return [status, (y - model) / err]

    else:
        return [status, y - model]


def gfit1d(y, x=None, err=None, weights=None, par=None, parinfo=None,
           maxiter=200, quiet=0):
    """
    Return the gaussian fit as an object.

    Parameters
    ----------
    y:   1D Numpy array
        The data to be fitted
    x:   1D Numpy array
        (optional) The x values of the y array. x and y must
        have the same shape.
    err: 1D Numpy array
        (optional) 1D array with measurement errors, must be
        the same shape as y
    weights: 1D Numpy array
        (optiional) 1D array with weights, must be the same
        shape as y
    par:  List
        (optional) Starting values for the parameters to be fitted
    parinfo: Dictionary of lists
        (optional) provides additional information for the
        parameters. For a detailed description see nmpfit.py.
        Parinfo can be used to limit parameters or keep
        some of them fixed.
    maxiter: number
        Maximum number of iterations to perform
        Default: 200
    quiet: number
        if set to 1, nmpfit does not print to the screen
        Default: 0

    Examples
    --------
    >>> x = np.arange(10,20, 0.1)
    >>> y= 10*np.e**(-(x-15)**2/4)
    >>> print(gfit1d(y,x=x, maxiter=20,quiet=1).params)
    [10.         15.          1.41421356]

    """
    y = y.astype(float)
    if weights is not None:
        weights = weights.astype(float)
    if err is not None:
        err = err.astype(float)
    if x is None and len(y.shape) == 1:
        x = np.arange(len(y)).astype(float)
    if x.shape != y.shape:
        print("input arrays X and Y must be of equal shape.\n")
        return

    fa = {'x': x, 'y': y, 'err': err, 'weights': weights}

    if par is not None:
        p = par
    else:
        ysigma = y.std()
        ind = np.nonzero(y > ysigma)[0]
        if len(ind) != 0:
            xind = int(ind.mean())
            p2 = x[xind]
            p1 = y[xind]
            p3 = 1.0
        else:
            ymax = y.max()
            ymin = y.min()
            ymean= y.mean()
            if (ymax - ymean) > (abs(ymin - ymean)):
                p1 = ymax
            else: p1 = ymin
            ind = (np.nonzero(y == p1))[0]
            p2 = x.mean()
            p3 = 1.


        p = [p1, p2, p3]
    m = nmpfit.mpfit(_gauss_funct, p,parinfo = parinfo, functkw=fa,
    maxiter=maxiter, quiet=quiet)
    if (m.status <= 0): print('error message = ', m.errmsg)
    return m


def plot_fit(y, mfit, x=None):
    if x is None:
        x = N.arange(len(y))
    else:
        x = x
    p = mfit.params
    #y = gauss_funct(p, y)
    yy = p[0] + N.e**(-0.5*(x-p[1])**2/p[2]**2)
    try:
        import pylab
    except ImportError:
        print("Matplotlib is not available.\n")
        return
    pylab.plot(x,yy)



"""A module that provides functions for manipulating bitmasks and data quality
(DQ) arrays.

"""
import warnings
import numpy as np
from astropy.utils import deprecated

__version__ = '1.1.2'
__vdate__ = '10-August-2021'
__author__ = 'Mihai Cara'

__all__ = ['interpret_bit_flags', 'bitfield_to_boolean_mask', 'is_bit_flag']

# Revision history:
# 0.1.0 (29-March-2015) - initial release based on code from stsci.skypac
# 0.1.1 (21-February-2017) - documentation typo fix
# 0.2.0 (23-February-2017) - performance and stability improvements. Changed
#       default output mask type from numpy.uint8 to numpy.bool_.
# 1.0.0 (16-March-2017) - Multiple enhancements:
#       1. Deprecated 'interpret_bits_value()'in favor of
#          'interpret_bit_flags()' which now takes 'flip_bits' argument to flip
#          bits in (list of) integer flags.
#       2. Deprecated 'bitmask2mask()' in favor of 'bitfield_to_boolean_mask()'
#          which now also takes 'flip_bits' argument.
#       3. Renamed arguments of 'interpret_bit_flags()' and
#          'bitfield_to_boolean_mask()' to be more technically correct.
#       4. 'interpret_bit_flags()' and 'bitfield_to_boolean_mask()' now
#          accept Python lists of bit flags (in addition to integer bitmasks
#          and string comma- (or '+') separated lists of bit flags).
#       5. Added 'is_bit_flag()' function to check if an integer number has
#          only one bit set (i.e., that it is a power of 2).
# 1.1.0 (29-January-2018) - Multiple enhancements:
#       1. Added support for long type in Python 2.7 in
#          `interpret_bit_flags()` and `bitfield_to_boolean_mask()`.
#       2. `interpret_bit_flags()` now always returns `int` (or `int` or `long`
#           in Python 2.7). Previously when input was of integer-like type
#           (i.e., `numpy.uint64`), it was not converted to Python `int`.
#       3. `bitfield_to_boolean_mask()` will no longer crash when
#          `ignore_flags` argument contains bit flags beyond what the type of
#          the argument `bitfield` can hold.
# 1.1.1 (30-January-2018) - Improved filtering of high bits in flags.
# 1.1.2 (11-August-2021) - Improved compatibility with numpy 1.21.*.
#
INT_TYPE = (int, )
MAX_UINT_TYPE = np.maximum_sctype(np.uint)
SUPPORTED_FLAGS = int(np.bitwise_not(
    0, dtype=MAX_UINT_TYPE, casting='unsafe'
))


def is_bit_flag(n):
    """
    Verifies if the input number is a bit flag (i.e., an integer number that is
    an integer power of 2).

    Parameters
    ----------
    n : int
        A positive integer number. Non-positive integers are considered not to
        be "flags".

    Returns
    -------
    result : bool
        ``True`` if input ``n`` is a bit flag and ``False`` if it is not.

    """
    if n < 1:
        return False

    return bin(n).count('1') == 1


def _is_int(n):
    return (
        (isinstance(n, INT_TYPE) and not isinstance(n, bool)) or
        (isinstance(n, np.generic) and np.issubdtype(n, np.integer))
    )


def interpret_bit_flags(bit_flags, flip_bits=None):
    """
    Converts input bit flags to a single integer value (bitmask) or `None`.

    When input is a list of flags (either a Python list of integer flags or a
    sting of comma- or '+'-separated list of flags), the returned bitmask
    is obtained by summing input flags.

    .. note::
        In order to flip the bits of the returned bitmask,
        for input of `str` type, prepend '~' to the input string. '~' must
        be prepended to the *entire string* and not to each bit flag! For
        input that is already a bitmask or a Python list of bit flags, set
        `flip_bits` for `True` in order to flip the bits of the returned
        bitmask.

    Parameters
    ----------
    bit_flags : int, str, list, None
        An integer bitmask or flag, `None`, a string of comma- or
        '+'-separated list of integer bit flags, or a Python list of integer
        bit flags. If `bit_flags` is a `str` and if it is prepended with '~',
        then the output bitmask will have its bits flipped (compared to simple
        sum of input flags). For input `bit_flags` that is already a bitmask
        or a Python list of bit flags, bit-flipping can be controlled through
        `flip_bits` parameter.

    flip_bits : bool, None
        Indicates whether or not to flip the bits of the returned bitmask
        obtained from input bit flags. This parameter must be set to `None`
        when input `bit_flags` is either `None` or a Python list of flags.

    Returns
    -------
    bitmask : int or None
        Returns and integer bit mask formed from the input bit value
        or `None` if input `bit_flags` parameter is `None` or an empty string.
        If input string value was prepended with '~' (or `flip_bits` was
        set to `True`), then returned value will have its bits flipped
        (inverse mask).

    Examples
    --------
    >>> from stsci.tools.bitmask import interpret_bit_flags
    >>> "{0:016b}".format(0xFFFF & interpret_bit_flags(28))
    '0000000000011100'
    >>> "{0:016b}".format(0xFFFF & interpret_bit_flags('4,8,16'))
    '0000000000011100'
    >>> "{0:016b}".format(0xFFFF & interpret_bit_flags('~4,8,16'))
    '1111111111100011'
    >>> "{0:016b}".format(0xFFFF & interpret_bit_flags('~(4+8+16)'))
    '1111111111100011'
    >>> "{0:016b}".format(0xFFFF & interpret_bit_flags([4, 8, 16]))
    '0000000000011100'
    >>> "{0:016b}".format(0xFFFF & interpret_bit_flags([4, 8, 16], flip_bits=True))
    '1111111111100011'

    """
    has_flip_bits = flip_bits is not None
    flip_bits = bool(flip_bits)
    allow_non_flags = False

    if _is_int(bit_flags):
        return (~int(bit_flags) if flip_bits else int(bit_flags))

    elif bit_flags is None:
        if has_flip_bits:
            raise TypeError(
                "Keyword argument 'flip_bits' must be set to 'None' when "
                "input 'bit_flags' is None."
            )
        return None

    elif isinstance(bit_flags, str):
        if has_flip_bits:
            raise TypeError(
                "Keyword argument 'flip_bits' is not permitted for "
                "comma-separated string lists of bit flags. Prepend '~' to "
                "the string to indicate bit-flipping."
            )

        bit_flags = str(bit_flags).strip()

        if bit_flags.upper() in ['', 'NONE', 'INDEF']:
            return None

        # check whether bitwise-NOT is present and if it is, check that it is
        # in the first position:
        bitflip_pos = bit_flags.find('~')
        if bitflip_pos == 0:
            flip_bits = True
            bit_flags = bit_flags[1:].lstrip()
        else:
            if bitflip_pos > 0:
                raise ValueError("Bitwise-NOT must precede bit flag list.")
            flip_bits = False

        # basic check for correct use of parenthesis:
        while True:
            nlpar = bit_flags.count('(')
            nrpar = bit_flags.count(')')

            if nlpar == 0 and nrpar == 0:
                break

            if nlpar != nrpar:
                raise ValueError("Unbalanced parantheses in bit flag list.")

            lpar_pos = bit_flags.find('(')
            rpar_pos = bit_flags.rfind(')')
            if lpar_pos > 0 or rpar_pos < (len(bit_flags) - 1):
                raise ValueError("Incorrect syntax (incorrect use of "
                                 "parenthesis) in bit flag list.")

            bit_flags = bit_flags[1:-1].strip()

        if ',' in bit_flags:
            bit_flags = bit_flags.split(',')

        elif '+' in bit_flags:
            bit_flags = bit_flags.split('+')

        else:
            if bit_flags == '':
                raise ValueError(
                    "Empty bit flag lists not allowed when either bitwise-NOT "
                    "or parenthesis are present."
                )
            bit_flags = [bit_flags]

        allow_non_flags = len(bit_flags) == 1

    elif hasattr(bit_flags, '__iter__'):
        if not all([_is_int(flag) for flag in bit_flags]):
            raise TypeError("Each bit flag in a list must be an integer.")

    else:
        raise TypeError("Unsupported type for argument 'bit_flags'.")

    bitset = set(map(int, bit_flags))
    if len(bitset) != len(bit_flags):
        warnings.warn("Duplicate bit flags will be ignored")

    bitmask = 0
    for v in bitset:
        if not is_bit_flag(v) and not allow_non_flags:
            raise ValueError("Input list contains invalid (not powers of two) "
                             "bit flags")
        bitmask += v

    if flip_bits:
        bitmask = ~bitmask

    return bitmask


def bitfield_to_boolean_mask(bitfield, ignore_flags=0, flip_bits=None,
                             good_mask_value=True, dtype=np.bool_):
    """
    Converts an array of bit fields to a boolean (or integer) mask array
    according to a bitmask constructed from the supplied bit flags (see
    ``ignore_flags`` parameter).

    This function is particularly useful to convert data quality arrays to
    boolean masks with selective filtering of DQ flags.

    Parameters
    ----------
    bitfield : numpy.ndarray
        An array of bit flags. By default, values different from zero are
        interpreted as "bad" values and values equal to zero are considered
        as "good" values. However, see ``ignore_flags`` parameter on how to
        selectively ignore some bits in the ``bitfield`` array data.

    ignore_flags : int, str, list, None (Default = 0)
        An integer bitmask, a Python list of bit flags, a comma- or
        '+'-separated string list of integer bit flags that indicate what
        bits in the input ``bitfield`` should be *ignored* (i.e., zeroed), or
        `None`.

        Setting ``ignore_flags`` to `None` effectively will make
        `bitfield_to_boolean_mask` interpret all ``bitfield`` elements
        as "good" regardless of their value.

        When ``ignore_flags`` argument is an integer bitmask, it will be
        combined using bitwise-NOT and bitwise-AND with each element of the
        input ``bitfield`` array (``~ignore_flags & bitfield``). If the
        resultant bitfield element is non-zero, that element will be
        interpreted as a "bad" in the output boolean mask and it will be
        interpreted as "good" otherwise. ``flip_bits`` parameter may be used
        to flip the bits (``bitwise-NOT``) of the bitmask thus effectively
        changing the meaning of the ``ignore_flags`` parameter from "ignore"
        to "use only" these flags.

        .. note::

            Setting ``ignore_flags`` to 0 effectively will assume that all
            non-zero elements in the input ``bitfield`` array are to be
            interpreted as "bad".

        When ``ignore_flags`` argument is an Python list of integer bit
        flags, these flags are added together to create an integer bitmask.
        Each item in the list must be a flag, i.e., an integer that is an
        integer power of 2. In order to flip the bits of the resultant
        bitmask, use ``flip_bits`` parameter.

        Alternatively, ``ignore_flags`` may be a string of comma- or
        '+'-separated list of integer bit flags that should be added together
        to create an integer bitmask. For example, both ``'4,8'`` and
        ``'4+8'`` are equivalent and indicate that bit flags 4 and 8 in
        the input ``bitfield`` array should be ignored when generating
        boolean mask.

        .. note::

            ``'None'``, ``'INDEF'``, and empty (or all white space) strings
            are special values of string ``ignore_flags`` that are
            interpreted as `None`.

        .. note::

            Each item in the list must be a flag, i.e., an integer that is an
            integer power of 2. In addition, for convenience, an arbitrary
            **single** integer is allowed and it will be interpretted as an
            integer bitmask. For example, instead of ``'4,8'`` one could
            simply provide string ``'12'``.

        .. note::

            When ``ignore_flags`` is a `str` and when it is prepended with
            '~', then the meaning of ``ignore_flags`` parameters will be
            reversed: now it will be interpreted as a list of bit flags to be
            *used* (or *not ignored*) when deciding which elements of the
            input ``bitfield`` array are "bad". Following this convention,
            an ``ignore_flags`` string value of ``'~0'`` would be equivalent
            to setting ``ignore_flags=None``.

        .. warning::

            Because prepending '~' to a string ``ignore_flags`` is equivalent
            to setting ``flip_bits`` to `True`, ``flip_bits`` cannot be used
            with string ``ignore_flags`` and it must be set to `None`.

    flip_bits : bool, None (Default = None)
        Specifies whether or not to invert the bits of the bitmask either
        supplied directly through ``ignore_flags`` parameter or built from the
        bit flags passed through ``ignore_flags`` (only when bit flags are
        passed as Python lists of integer bit flags). Occasionally, it may be
        useful to *consider only specific bit flags* in the ``bitfield``
        array when creating a boolean mask as opposite to *ignoring* specific
        bit flags as ``ignore_flags`` behaves by default. This can be achieved
        by inverting/flipping the bits of the bitmask created from
        ``ignore_flags`` flags which effectively changes the meaning of the
        ``ignore_flags`` parameter from "ignore" to "use only" these flags.
        Setting ``flip_bits`` to `None` means that no bit flipping will be
        performed. Bit flipping for string lists of bit flags must be
        specified by prepending '~' to string bit flag lists
        (see documentation for ``ignore_flags`` for more details).

        .. warning::
            This parameter can be set to either `True` or `False` **ONLY** when
            ``ignore_flags`` is either an integer bitmask or a Python
            list of integer bit flags. When ``ignore_flags`` is either
            `None` or a string list of flags, ``flip_bits`` **MUST** be set
            to `None`.

    good_mask_value : int, bool (Default = True)
        This parameter is used to derive the values that will be assigned to
        the elements in the output boolean mask array that correspond to the
        "good" bit fields (that are 0 after zeroing bits specified by
        ``ignore_flags``) in the input ``bitfield`` array. When
        ``good_mask_value`` is non-zero or `True` then values in the output
        boolean mask array corresponding to "good" bit fields in ``bitfield``
        will be `True` (if ``dtype`` is `numpy.bool_`) or 1 (if ``dtype`` is
        of numerical type) and values of corresponding to "bad" flags will be
        `False` (or 0). When ``good_mask_value`` is zero or `False` then the
        values in the output boolean mask array corresponding to "good" bit
        fields in ``bitfield`` will be `False` (if ``dtype`` is `numpy.bool_`)
        or 0 (if ``dtype`` is of numerical type) and values of corresponding
        to "bad" flags will be `True` (or 1).

    dtype : data-type
        The desired data-type for the output binary mask array.

    Returns
    -------
    mask : numpy.ndarray
        Returns an array of the same dimensionality as the input ``bitfield``
        array whose elements can have two possible values,
        e.g., `True` or `False` (or 1 or 0 for integer ``dtype``) according to
        values of to the input ``bitfield`` elements, ``ignore_flags``
        parameter, and the ``good_mask_value`` parameter.

    Examples
    --------
    >>> from stsci.tools import bitmask
    >>> import numpy as np
    >>> dqbits = np.asarray([[0,0,1,2,0,8,12,0],[10,4,0,0,0,16,6,0]])
    >>> bitmask.bitfield_to_boolean_mask(dqbits, ignore_flags=0, dtype=int)
    array([[1, 1, 0, 0, 1, 0, 0, 1],
           [0, 0, 1, 1, 1, 0, 0, 1]])
    >>> bitmask.bitfield_to_boolean_mask(dqbits, ignore_flags=0, dtype=bool)
    array([[ True,  True, False, False,  True, False, False,  True],
           [False, False,  True,  True,  True, False, False,  True]])
    >>> bitmask.bitfield_to_boolean_mask(dqbits, ignore_flags=6, good_mask_value=0, dtype=int)
    array([[0, 0, 1, 0, 0, 1, 1, 0],
           [1, 0, 0, 0, 0, 1, 0, 0]])
    >>> bitmask.bitfield_to_boolean_mask(dqbits, ignore_flags=~6, good_mask_value=0, dtype=int)
    array([[0, 0, 0, 1, 0, 0, 1, 0],
           [1, 1, 0, 0, 0, 0, 1, 0]])
    >>> bitmask.bitfield_to_boolean_mask(dqbits, ignore_flags=6, flip_bits=True, good_mask_value=0, dtype=int)
    array([[0, 0, 0, 1, 0, 0, 1, 0],
           [1, 1, 0, 0, 0, 0, 1, 0]])
    >>> bitmask.bitfield_to_boolean_mask(dqbits, ignore_flags='~(2+4)', good_mask_value=0, dtype=int)
    array([[0, 0, 0, 1, 0, 0, 1, 0],
           [1, 1, 0, 0, 0, 0, 1, 0]])
    >>> bitmask.bitfield_to_boolean_mask(dqbits, ignore_flags=[2, 4], flip_bits=True, good_mask_value=0, dtype=int)
    array([[0, 0, 0, 1, 0, 0, 1, 0],
           [1, 1, 0, 0, 0, 0, 1, 0]])

    """
    bitfield = np.asarray(bitfield)
    if not np.issubdtype(bitfield.dtype, np.integer):
        raise TypeError("Input bitfield array must be of integer type.")

    ignore_mask = interpret_bit_flags(ignore_flags, flip_bits=flip_bits)

    if ignore_mask is None:
        if good_mask_value:
            mask = np.ones_like(bitfield, dtype=dtype)
        else:
            mask = np.zeros_like(bitfield, dtype=dtype)
        return mask

    # filter out bits beyond the maximum supported by the data type:
    ignore_mask = ignore_mask & SUPPORTED_FLAGS

    # invert the "ignore" mask:
    ignore_mask = np.bitwise_not(ignore_mask, dtype=bitfield.dtype.type,
                                 casting='unsafe')

    mask = np.empty_like(bitfield, dtype=np.bool_)
    np.bitwise_and(bitfield, ignore_mask, out=mask, casting='unsafe')

    if good_mask_value:
        np.logical_not(mask, out=mask)

    return mask.astype(dtype=dtype, subok=False, copy=False)


@deprecated(since='3.4.6', message='', name='interpret_bits_value',
            alternative='interpret_bit_flags')
def interpret_bits_value(val):
    """
    Converts input bits value from string to a single integer value or None.
    If a comma- or '+'-separated set of values are provided, they are summed.

    .. note::
        In order to flip the bits of the final result (after summation),
        for input of `str` type, prepend '~' to the input string. '~' must
        be prepended to the *entire string* and not to each bit flag!

    Parameters
    ----------
    val : int, str, None
        An integer bit mask or flag, `None`, or a comma- or '+'-separated
        string list of integer bit values. If `val` is a `str` and if
        it is prepended with '~', then the output bit mask will have its
        bits flipped (compared to simple sum of input val).

    Returns
    -------
    bitmask : int or None
        Returns and integer bit mask formed from the input bit value
        or `None` if input `val` parameter is `None` or an empty string.
        If input string value was prepended with '~', then returned
        value will have its bits flipped (inverse mask).

    """
    if isinstance(val, int) or val is None:
        return val

    else:
        val = str(val).strip()

        if val.startswith('~'):
            flip_bits = True
            val = val[1:].lstrip()
        else:
            flip_bits = False

        if val.startswith('('):
            if val.endswith(')'):
                val = val[1:-1].strip()
            else:
                raise ValueError('Unbalanced parantheses or incorrect syntax.')

        if ',' in val:
            valspl = val.split(',')
            bitmask = 0
            for v in valspl:
                bitmask += int(v)

        elif '+' in val:
            valspl = val.split('+')
            bitmask = 0
            for v in valspl:
                bitmask += int(v)

        elif val.upper() in ['', 'NONE', 'INDEF']:
            return None

        else:
            bitmask = int(val)

        if flip_bits:
            bitmask = ~bitmask

    return bitmask


@deprecated(since='3.4.6', message='', name='bitmask2mask',
            alternative='bitfield_to_boolean_mask')
def bitmask2mask(bitmask, ignore_bits, good_mask_value=1, dtype=np.bool_):
    r"""
    bitmask2mask(bitmask, ignore_bits, good_mask_value=1, dtype=numpy.bool\_)
    Interprets an array of bit flags and converts it to a "binary" mask array.
    This function is particularly useful to convert data quality arrays to
    binary masks.

    Parameters
    ----------
    bitmask : numpy.ndarray
        An array of bit flags. Values different from zero are interpreted as
        "bad" values and values equal to zero are considered as "good" values.
        However, see `ignore_bits` parameter on how to ignore some bits
        in the `bitmask` array.

    ignore_bits : int, str, None
        An integer bit mask, `None`, or a comma- or '+'-separated
        string list of integer bit values that indicate what bits in the
        input `bitmask` should be *ignored* (i.e., zeroed). If `ignore_bits`
        is a `str` and if it is prepended with '~', then the meaning
        of `ignore_bits` parameters will be reversed: now it will be
        interpreted as a list of bits to be *used* (or *not ignored*) when
        deciding what elements of the input `bitmask` array are "bad".

        The `ignore_bits` parameter is the integer sum of all of the bit
        values from the input `bitmask` array that should be considered
        "good" when creating the output binary mask. For example, if
        values in the `bitmask` array can be combinations
        of 1, 2, 4, and 8 flags and one wants to consider that
        values having *only* bit flags 2 and/or 4 as being "good",
        then `ignore_bits` should be set to 2+4=6. Then a `bitmask` element
        having values 2,4, or 6 will be considered "good", while an
        element with a value, e.g., 1+2=3, 4+8=12, etc. will be interpreted
        as "bad".

        Alternatively, one can enter a comma- or '+'-separated list
        of integer bit flags that should be added to obtain the
        final "good" bits. For example, both ``4,8`` and ``4+8``
        are equivalent to setting `ignore_bits` to 12.

        See :py:func:`interpret_bits_value` for examples.

        | Setting `ignore_bits` to `None` effectively will interpret
          all `bitmask` elements as "good" regardless of their value.

        | Setting `ignore_bits` to 0 effectively will assume that all
          non-zero elements in the input `bitmask` array are to be
          interpreted as "bad".

        | In order to reverse the meaning of the `ignore_bits`
          parameter from indicating bits in the values of `bitmask`
          elements that should be ignored when deciding which elements
          are "good" (these are the elements that are zero after ignoring
          `ignore_bits`), to indicating the bits should be used
          exclusively in deciding whether a `bitmask` element is "good",
          prepend '~' to the string value. For example, in order to use
          **only** (or **exclusively**) flags 4 and 8 (2nd and 3rd bits)
          in the values of the input `bitmask` array when deciding whether
          or not that element is "good", set `ignore_bits` to ``~4+8``,
          or ``~4,8 To obtain the same effect with an `int` input value
          (except for 0), enter -(4+8+1)=-9. Following this convention,
          a `ignore_bits` string value of ``'~0'`` would be equivalent to
          setting ``ignore_bits=None``.

    good_mask_value : int, bool (Default = 1)
        This parameter is used to derive the values that will be assigned to
        the elements in the output `mask` array that correspond to the "good"
        flags (that are 0 after zeroing bits specified by `ignore_bits`)
        in the input `bitmask` array. When `good_mask_value` is non-zero or
        `True` then values in the output mask array corresponding to "good"
        bit flags in `bitmask` will be 1 (or `True` if `dtype` is `bool`) and
        values of corresponding to "bad" flags will be 0. When
        `good_mask_value` is zero or `False` then values in the output mask
        array corresponding to "good" bit flags in `bitmask` will be 0
        (or `False` if `dtype` is `bool`) and values of corresponding
        to "bad" flags will be 1.

    dtype : data-type (Default = numpy.uint8)
        The desired data-type for the output binary mask array.

    Returns
    -------
    mask : numpy.ndarray
        Returns an array whose elements can have two possible values,
        e.g., 1 or 0 (or `True` or `False` if `dtype` is `bool`) according to
        values of to the input `bitmask` elements, `ignore_bits` parameter,
        and the `good_mask_value` parameter.

    """
    if not np.issubdtype(bitmask.dtype, np.integer):
        raise TypeError("Input 'bitmask' array must be of integer type.")

    ignore_bits = interpret_bits_value(ignore_bits)

    if ignore_bits is None:
        if good_mask_value:
            mask = np.ones_like(bitmask, dtype=dtype)
        else:
            mask = np.zeros_like(bitmask, dtype=dtype)
        return mask

    ignore_bits = ~bitmask.dtype.type(ignore_bits)

    mask = np.empty_like(bitmask, dtype=np.bool_)
    np.bitwise_and(bitmask, ignore_bits, out=mask, casting='unsafe')

    if good_mask_value:
        np.logical_not(mask, out=mask)

    return mask.astype(dtype=dtype, subok=False, copy=False)



#!/usr/bin/env python
import os
import tempfile

import numpy as np
import pytest
from astropy.io import fits
from astropy.io.fits.tests import FitsTestCase

from stsci.tools import stpyfits


class TestStpyfitsFunctions(FitsTestCase):
    def setup(self):
        self.data_dir = os.path.join(os.path.dirname(__file__), 'data')
        self.temp_dir = tempfile.mkdtemp(prefix='stpyfits-test-')
        self.writekwargs = {'overwrite': True}

    def test_InfoConvienceFunction(self):
        """Test the info convience function in both the fits and stpyfits
        namespace."""

        ans1 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 215, (), '', ''),
                (1, 'SCI', 1, 'ImageHDU', 141, (62, 44), 'int16 (rescales to uint16)', ''),
                (2, 'ERR', 1, 'ImageHDU', 71, (62, 44), 'int16', ''),
                (3, 'DQ', 1, 'ImageHDU', 71, (62, 44), 'int16', ''),
                (4, 'SCI', 2, 'ImageHDU', 141, (62, 44), 'int16 (rescales to uint16)', ''),
                (5, 'ERR', 2, 'ImageHDU', 71, (62, 44), 'int16', ''),
                (6, 'DQ', 2, 'ImageHDU', 71, (62, 44), 'int16', '')]
        ans2 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 215, (), '', ''),
                (1, 'SCI', 1, 'ImageHDU', 141, (62, 44), 'int16 (rescales to uint16)', ''),
                (2, 'ERR', 1, 'ImageHDU', 71, (), '', ''),
                (3, 'DQ', 1, 'ImageHDU', 71, (), '', ''),
                (4, 'SCI', 2, 'ImageHDU', 141, (62, 44), 'int16 (rescales to uint16)', ''),
                (5, 'ERR', 2, 'ImageHDU', 71, (), '', ''),
                (6, 'DQ', 2, 'ImageHDU', 71, (), '', '')]
        ans3 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (10, 10), 'int32', '')]
        ans4 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (), '', '')]

        assert stpyfits.info(self.data('o4sp040b0_raw.fits'), output=False) == ans1
        assert fits.info(self.data('o4sp040b0_raw.fits'), output=False) == ans2

        assert stpyfits.info(self.data('cdva2.fits'), output=False) == ans3
        assert fits.info(self.data('cdva2.fits'), output=False) == ans4

    def test_OpenConvienceFunction(self):
        """Test the open convience function in both the fits and stpyfits
        namespace."""

        hdul = stpyfits.open(self.data('cdva2.fits'))
        hdul1 = fits.open(self.data('cdva2.fits'))

        assert hdul[0].header['NAXIS'] == 2
        assert hdul1[0].header['NAXIS'] == 0
        assert hdul[0].header['NAXIS1'] == 10
        assert hdul[0].header['NAXIS2'] == 10

        for k in ('NAXIS1', 'NAXIS2'):
            with pytest.raises(KeyError):
                hdul1[0].header[k]
        for k in ('NPIX1', 'NPIX2'):
            with pytest.raises(KeyError):
                hdul[0].header[k]

        assert hdul1[0].header['NPIX1'] == 10
        assert hdul1[0].header['NPIX2'] == 10

        assert (hdul[0].data == np.ones((10, 10), dtype=np.int32)).all()

        assert hdul1[0].data is None

        hdul.close()
        hdul1.close()

    @pytest.mark.parametrize(['filename', 'ext', 'naxis1', 'naxis2'],
                             [('cdva2.fits', None, 10, 10),
                              ('o4sp040b0_raw.fits', 2, 62, 44)])
    def test_GetHeaderConvienceFunction(self, filename, ext, naxis1, naxis2):
        """Test the getheader convience function in both the fits and
        stpyfits namespace."""

        if ext is None:
            hd = stpyfits.getheader(self.data(filename))
            hd1 = fits.getheader(self.data(filename))
        else:
            hd = stpyfits.getheader(self.data(filename), ext)
            hd1 = fits.getheader(self.data(filename), ext)

        assert hd['NAXIS'] == 2
        assert hd1['NAXIS'] == 0
        assert hd['NAXIS1'] == naxis1
        assert hd['NAXIS2'] == naxis2

        for k in ('NAXIS1', 'NAXIS2'):
            with pytest.raises(KeyError):
                hd1[k]
        for k in ('NPIX1', 'NPIX2'):
            with pytest.raises(KeyError):
                hd[k]

        assert hd1['NPIX1'] == naxis1
        assert hd1['NPIX2'] == naxis2

    def test_GetDataConvienceFunction(self):
        """Test the getdata convience function in both the fits and
        stpyfits namespace."""

        d = stpyfits.getdata(self.data('cdva2.fits'))
        assert (d == np.ones((10, 10), dtype=np.int32)).all()
        with pytest.raises(IndexError):
            fits.getdata(self.data('cdva2.fits'))

    def test_GetValConvienceFunction(self):
        """Test the getval convience function in both the fits and
        stpyfits namespace."""

        val = stpyfits.getval(self.data('cdva2.fits'), 'NAXIS', 0)
        val1 = fits.getval(self.data('cdva2.fits'), 'NAXIS', 0)
        assert val == 2
        assert val1 == 0

    def test_writetoConvienceFunction(self):
        """Test the writeto convience function in both the fits and stpyfits
        namespace."""

        hdul = stpyfits.open(self.data('cdva2.fits'))
        hdul1 = fits.open(self.data('cdva2.fits'))

        header = hdul[0].header.copy()
        header['NAXIS'] = 0

        stpyfits.writeto(self.temp('new.fits'), hdul[0].data, header,
                         **self.writekwargs)
        fits.writeto(self.temp('new1.fits'), hdul1[0].data, hdul1[0].header,
                     **self.writekwargs)

        hdul.close()
        hdul1.close()

        info1 = fits.info(self.temp('new.fits'), output=False)
        info2 = stpyfits.info(self.temp('new.fits'), output=False)
        info3 = fits.info(self.temp('new1.fits'), output=False)
        info4 = stpyfits.info(self.temp('new1.fits'), output=False)

        ans1 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 6, (), '', '')]
        ans2 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 6, (10, 10), 'int32', '')]
        ans3 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 6, (), '', '')]
        ans4 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 6, (10, 10), 'uint8', '')]

        assert info1 == ans1
        assert info2 == ans2
        assert info3 == ans3
        assert info4 == ans4

    def test_appendConvienceFunction(self):
        """Test the append convience function in both the fits and stpyfits
        namespace."""

        hdul = stpyfits.open(self.data('cdva2.fits'))
        hdul1 = fits.open(self.data('cdva2.fits'))

        stpyfits.writeto(self.temp('new.fits'), hdul[0].data, hdul[0].header,
                         **self.writekwargs)
        fits.writeto(self.temp('new1.fits'), hdul1[0].data, hdul1[0].header,
                     **self.writekwargs)

        hdu = stpyfits.ImageHDU()
        hdu1 = fits.ImageHDU()

        hdu.data = hdul[0].data
        hdu1.data = hdul1[0].data
        hdu.header.set('BITPIX', 32)
        hdu1.header.set('BITPIX', 32)
        hdu.header.set('NAXIS', 2)
        hdu.header.set('NAXIS1', 10, 'length of constant array axis 1',
                       after='NAXIS')
        hdu.header.set('NAXIS2', 10, 'length of constant array axis 2',
                       after='NAXIS1')
        hdu.header.set('PIXVALUE', 1, 'Constant pixel value', after='GCOUNT')
        hdu1.header.set('PIXVALUE', 1, 'Constant pixel value', after='GCOUNT')
        hdu1.header.set('NPIX1', 10, 'length of constant array axis 1',
                        after='GCOUNT')
        hdu1.header.set('NPIX2', 10, 'length of constant array axis 2',
                        after='NPIX1')
        stpyfits.append(self.temp('new.fits'), hdu.data, hdu.header)
        fits.append(self.temp('new1.fits'), hdu1.data, hdu1.header)

        ans1 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (10, 10), 'int32', ''),
                (1, '', 1, 'ImageHDU', 8, (10, 10), 'int32', '')]
        ans2 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (10, 10), 'uint8', ''),
                (1, '', 1, 'ImageHDU', 8, (10, 10), 'uint8', '')]
        ans3 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (10, 10), 'int32', ''),
                (1, '', 1, 'ImageHDU', 8, (10, 10), 'int32', '')]
        ans4 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (), '', ''),
                (1, '', 1, 'ImageHDU', 8, (), '', '')]

        assert stpyfits.info(self.temp('new.fits'), output=False) == ans1
        assert stpyfits.info(self.temp('new1.fits'), output=False) == ans2
        assert fits.info(self.temp('new.fits'), output=False) == ans3
        assert fits.info(self.temp('new1.fits'), output=False) == ans4

        hdul5 = stpyfits.open(self.temp('new.fits'))
        hdul6 = fits.open(self.temp('new1.fits'))

        assert hdul5[1].header['NAXIS'] == 2
        assert hdul6[1].header['NAXIS'] == 0
        assert hdul5[1].header['NAXIS1'] == 10
        assert hdul5[1].header['NAXIS2'] == 10

        for k in ('NPIX1', 'NPIX2'):
            with pytest.raises(KeyError):
                hdul5[1].header[k]
        for k in ('NAXIS1', 'NAXIS2'):
            with pytest.raises(KeyError):
                hdul6[1].header[k]

        assert hdul6[1].header['NPIX1'] == 10
        assert hdul6[1].header['NPIX2'] == 10
        assert (hdul5[1].data == np.ones((10, 10), dtype=np.int32)).all()
        assert hdul6[1].data is None

        hdul5.close()
        hdul6.close()
        hdul.close()
        hdul1.close()

    def test_updateConvienceFunction(self):
        """Test the update convience function in both the fits and stpyfits
        namespace."""

        hdul = stpyfits.open(self.data('cdva2.fits'))
        hdul1 = fits.open(self.data('cdva2.fits'))

        header = hdul[0].header.copy()
        header['NAXIS'] = 0

        stpyfits.writeto(self.temp('new.fits'), hdul[0].data, header,
                         **self.writekwargs)

        hdu = stpyfits.ImageHDU()
        hdu1 = fits.ImageHDU()

        hdu.data = hdul[0].data
        hdu1.data = hdul1[0].data
        hdu.header.set('BITPIX', 32)
        hdu1.header.set('BITPIX', 32)
        hdu.header.set('NAXIS', 0)
        hdu.header.set('PIXVALUE', 1, 'Constant pixel value', after='GCOUNT')
        hdu1.header.set('PIXVALUE', 1, 'Constant pixel value', after='GCOUNT')
        hdu.header.set('NPIX1', 10, 'length of constant array axis 1',
                       after='GCOUNT')
        hdu.header.set('NPIX2', 10, 'length of constant array axis 2',
                       after='NPIX1')
        stpyfits.append(self.temp('new.fits'), hdu.data, hdu.header)

        d = np.zeros_like(hdu.data)

        stpyfits.update(self.temp('new.fits'), d, hdu.header, 1)

        ans1 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (), '', ''),
                (1, '', 1, 'ImageHDU', 8, (), '', '')]
        ans2 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (10, 10), 'int32', ''),
                (1, '', 1, 'ImageHDU', 8, (10, 10), 'int32', '')]

        assert fits.info(self.temp('new.fits'), output=False) == ans1
        assert stpyfits.info(self.temp('new.fits'), output=False) == ans2

        hdul7 = stpyfits.open(self.temp('new.fits'))
        assert hdul7[1].header['NAXIS'] == 2
        assert hdul7[1].header['NAXIS1'] == 10
        assert hdul7[1].header['NAXIS2'] == 10
        assert hdul7[1].header['PIXVALUE'] == 0

        for k in ('NPIX1', 'NPIX2'):
            with pytest.raises(KeyError):
                hdul7[1].header[k]

        assert (hdul7[1].data == np.zeros((10, 10), dtype=np.int32)).all()

        hdul8 = fits.open(self.temp('new.fits'))
        assert hdul8[1].header['NAXIS'] == 0
        assert hdul8[1].header['NPIX1'] == 10
        assert hdul8[1].header['NPIX2'] == 10
        assert hdul8[1].header['PIXVALUE'] == 0

        for k in ('NAXIS1', 'NAXIS2'):
            with pytest.raises(KeyError):
                hdul8[1].header[k]

        assert hdul8[1].data is None

        hdul7.close()
        hdul8.close()
        hdul.close()
        hdul1.close()

    def test_ImageHDUConstructor(self):
        """Test the ImageHDU constructor in both the fits and stpyfits
        namespace."""

        hdu = stpyfits.ImageHDU()
        assert isinstance(hdu, stpyfits.ConstantValueImageHDU)
        assert isinstance(hdu, fits.ImageHDU)

    def test_PrimaryHDUConstructor(self):
        """Test the PrimaryHDU constructor in both the fits and stpyfits
        namespace.  Although stpyfits does not reimplement the
        constructor, it does add _ConstantValueImageBaseHDU to the
        inheritance hierarchy of fits.PrimaryHDU when accessed through the
        stpyfits namespace.  This method tests that that inheritance is
        working"""

        n = np.ones(10)

        hdu = stpyfits.PrimaryHDU(n)
        hdu.header.set('PIXVALUE', 1.0, 'Constant pixel value', after='EXTEND')
        hdu.header.set('NAXIS', 0)

        stpyfits.writeto(self.temp('new.fits'), hdu.data, hdu.header,
                         **self.writekwargs)

        hdul = stpyfits.open(self.temp('new.fits'))
        hdul1 = fits.open(self.temp('new.fits'))

        assert hdul[0].header['NAXIS'] == 1
        assert hdul[0].header['NAXIS1'] == 10
        assert hdul[0].header['PIXVALUE'] == 1.0

        with pytest.raises(KeyError):
            hdul[0].header['NPIX1']

        assert (hdul[0].data == np.ones(10, dtype=np.float32)).all()

        assert hdul1[0].header['NAXIS'] == 0
        assert hdul1[0].header['NPIX1'] == 10
        assert hdul1[0].header['PIXVALUE'] == 1.0

        with pytest.raises(KeyError):
            hdul1[0].header['NAXIS1']

        assert hdul1[0].data is None

        hdul.close()
        hdul1.close()

    def test_HDUListWritetoMethod(self):
        """Test the writeto method of HDUList in both the fits and stpyfits
           namespace."""

        hdu = stpyfits.PrimaryHDU()
        hdu1 = stpyfits.ImageHDU()
        hdu.data = np.zeros((10, 10), dtype=np.int32)
        hdu1.data = hdu.data + 2
        hdu.header.set('BITPIX', 32)
        hdu1.header.set('BITPIX', 32)
        hdu.header.set('NAXIS', 2)
        hdu.header.set('NAXIS1', 10, 'length of constant array axis 1',
                       after='NAXIS')
        hdu.header.set('NAXIS2', 10, 'length of constant array axis 2',
                       after='NAXIS1')
        hdu.header.set('PIXVALUE', 0, 'Constant pixel value')
        hdu1.header.set('PIXVALUE', 2, 'Constant pixel value', after='GCOUNT')
        hdu1.header.set('NAXIS', 2)
        hdu1.header.set('NAXIS1', 10, 'length of constant array axis 1',
                        after='NAXIS')
        hdu1.header.set('NAXIS2', 10, 'length of constant array axis 2',
                        after='NAXIS1')
        hdul = stpyfits.HDUList([hdu, hdu1])

        hdul.writeto(self.temp('new.fits'), **self.writekwargs)

        ans1 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (10, 10), 'int32', ''),
                (1, '', 1, 'ImageHDU', 8, (10, 10), 'int32', '')]
        ans2 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 7, (), '', ''),
                (1, '', 1, 'ImageHDU', 8, (), '', '')]

        assert stpyfits.info(self.temp('new.fits'), output=False) == ans1
        assert fits.info(self.temp('new.fits'), output=False) == ans2

        hdul1 = stpyfits.open(self.temp('new.fits'))
        hdul2 = fits.open(self.temp('new.fits'))

        assert hdul1[0].header['NAXIS'] == 2
        assert hdul1[0].header['NAXIS1'] == 10
        assert hdul1[0].header['NAXIS2'] == 10
        assert hdul1[0].header['PIXVALUE'] == 0

        assert (hdul1[0].data == np.zeros((10, 10), dtype=np.int32)).all()

        assert hdul1[1].header['NAXIS'] == 2
        assert hdul1[1].header['NAXIS1'] == 10
        assert hdul1[1].header['NAXIS2'] == 10
        assert hdul1[1].header['PIXVALUE'] == 2

        assert (hdul1[1].data == (np.zeros((10, 10), dtype=np.int32) + 2)).all()

        assert hdul2[0].header['NAXIS'] == 0
        assert hdul2[0].header['NPIX1'] == 10
        assert hdul2[0].header['NPIX2'] == 10
        assert hdul2[0].header['PIXVALUE'] == 0

        for i in range(2):
            for k in ('NPIX1', 'NPIX2'):
                with pytest.raises(KeyError):
                    hdul1[i].header[k]
            for k in ('NAXIS1', 'NAXIS2'):
                with pytest.raises(KeyError):
                    hdul2[i].header[k]

        assert hdul2[0].data is None

        assert hdul2[1].header['NAXIS'] == 0
        assert hdul2[1].header['NPIX1'] == 10
        assert hdul2[1].header['NPIX2'] == 10
        assert hdul2[1].header['PIXVALUE'] == 2

        hdul1.close()
        hdul2.close()

    def test_HDUList_getitem_Method(self):
        """Test the __getitem__ method of st_HDUList in the stpyfits
        namespace."""

        n = np.ones(10)

        hdu = stpyfits.PrimaryHDU(n)
        hdu.header.set('PIXVALUE', 1., 'constant pixel value', after='EXTEND')

        hdu.writeto(self.temp('new.fits'), **self.writekwargs)

        hdul = stpyfits.open(self.temp('new.fits'))
        hdul1 = fits.open(self.temp('new.fits'))

        hdu = hdul[0]
        hdu1 = hdul1[0]

        assert hdu.header['NAXIS'] == 1
        assert hdu.header['NAXIS1'] == 10
        assert hdu.header['PIXVALUE'] == 1.0

        with pytest.raises(KeyError):
            hdu.header['NPIX1']

        assert (hdu.data == np.ones(10, dtype=np.float32)).all()
        assert hdu1.header['NAXIS'] == 0
        assert hdu1.header['NPIX1'] == 10
        assert hdu1.header['PIXVALUE'] == 1.0

        with pytest.raises(KeyError):
            hdu1.header['NAXIS1']

        assert hdu1.data is None

        hdul.close()
        hdul1.close()

    def test_HDUListFlushMethod(self):
        """Test the flush method of HDUList in both the fits and stpyfits
        namespace."""

        hdu = stpyfits.PrimaryHDU()
        hdu1 = stpyfits.ImageHDU()
        hdu.data = np.zeros((10, 10), dtype=np.int32)
        hdu1.data = hdu.data + 2
        hdu.header.set('BITPIX', 32)
        hdu1.header.set('BITPIX', 32)
        hdu.header.set('NAXIS', 2)
        hdu.header.set('NAXIS1', 10, 'length of constant array axis 1',
                       after='NAXIS')
        hdu.header.set('NAXIS2', 10, 'length of constant array axis 2',
                       after='NAXIS1')
        hdu.header.set('PIXVALUE', 0, 'Constant pixel value')
        hdu1.header.set('PIXVALUE', 2, 'Constant pixel value', after='GCOUNT')
        hdu1.header.set('NAXIS', 2)
        hdu1.header.set('NAXIS1', 10, 'length of constant array axis 1',
                        after='NAXIS')
        hdu1.header.set('NAXIS2', 10, 'length of constant array axis 2',
                        after='NAXIS1')
        hdul = stpyfits.HDUList([hdu, hdu1])

        hdul.writeto(self.temp('new.fits'), **self.writekwargs)

        hdul = stpyfits.open(self.temp('new.fits'), 'update')
        hdul[0].data = np.zeros(10, dtype=np.int32) + 3
        hdul.flush()
        hdul.close()

        ans1 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 6, (10,), 'int32', ''),
                (1, '', 1, 'ImageHDU', 8, (10, 10), 'int32', '')]
        ans2 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 6, (), '', ''),
                (1, '', 1, 'ImageHDU', 8, (), '', '')]
        ans3 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 6, (15,), 'int32', ''),
                (1, '', 1, 'ImageHDU', 8, (10, 10), 'int32', '')]
        ans4 = [(0, 'PRIMARY', 1, 'PrimaryHDU', 6, (), '', ''),
                (1, '', 1, 'ImageHDU', 8, (), '', '')]

        assert stpyfits.info(self.temp('new.fits'), output=False) == ans1
        assert fits.info(self.temp('new.fits'), output=False) == ans2

        hdul1 = stpyfits.open(self.temp('new.fits'))
        hdul2 = fits.open(self.temp('new.fits'))

        assert hdul1[0].header['NAXIS'] == 1
        assert hdul1[0].header['NAXIS1'] == 10
        assert hdul1[0].header['PIXVALUE'] == 3

        with pytest.raises(KeyError):
            hdul1[0].header['NPIX1']

        assert (hdul1[0].data == (np.zeros(10, dtype=np.int32) + 3)).all()

        assert hdul2[0].header['NAXIS'] == 0
        assert hdul2[0].header['NPIX1'] == 10
        assert hdul2[0].header['PIXVALUE'] == 3

        with pytest.raises(KeyError):
            hdul2[0].header['NAXIS1']

        assert hdul2[0].data is None

        hdul1.close()
        hdul2.close()

        hdul3 = stpyfits.open(self.temp('new.fits'), 'update')
        hdul3[0].data = np.zeros(15, dtype=np.int32) + 4
        hdul3.close()  # Note that close calls flush

        assert stpyfits.info(self.temp('new.fits'), output=False) == ans3
        assert fits.info(self.temp('new.fits'), output=False) == ans4

        hdul1 = stpyfits.open(self.temp('new.fits'))
        hdul2 = fits.open(self.temp('new.fits'))

        assert hdul1[0].header['NAXIS'] == 1
        assert hdul1[0].header['NAXIS1'] == 15
        assert hdul1[0].header['PIXVALUE'] == 4

        with pytest.raises(KeyError):
            hdul1[0].header['NPIX1']

        assert (hdul1[0].data == (np.zeros(15, dtype=np.int32) + 4)).all()

        assert hdul2[0].header['NAXIS'] == 0
        assert hdul2[0].header['NPIX1'] == 15
        assert hdul2[0].header['PIXVALUE'] == 4

        with pytest.raises(KeyError):
            hdul2[0].header['NAXIS1']

        assert hdul2[0].data is None

        hdul1.close()
        hdul2.close()

    def test_ImageBaseHDU_getattr_Method(self):
        """Test the __getattr__ method of ImageBaseHDU in both the fits
        and stpyfits namespace."""

        hdul = stpyfits.open(self.data('cdva2.fits'))
        hdul1 = fits.open(self.data('cdva2.fits'))

        hdu = hdul[0]
        hdu1 = hdul1[0]

        assert (hdu.data == np.ones((10, 10), dtype=np.int32)).all()
        assert hdu1.data is None

        hdul.close()
        hdul1.close()

    def test_ImageBaseHDUWriteToMethod(self):
        """Test the writeto method of _ConstantValueImageBaseHDU in the
        stpyfits namespace."""

        n = np.ones(10)

        hdu = stpyfits.PrimaryHDU(n)
        hdu.header.set('PIXVALUE', 1., 'constant pixel value', after='EXTEND')

        hdu.writeto(self.temp('new.fits'), **self.writekwargs)

        hdul = stpyfits.open(self.temp('new.fits'))
        hdul1 = fits.open(self.temp('new.fits'))

        assert hdul[0].header['NAXIS'] == 1
        assert hdul[0].header['NAXIS1'] == 10
        assert hdul[0].header['PIXVALUE'] == 1.0

        with pytest.raises(KeyError):
            hdul[0].header['NPIX1']

        assert (hdul[0].data == np.ones(10, dtype=np.float32)).all()

        assert hdul1[0].header['NAXIS'] == 0
        assert hdul1[0].header['NPIX1'] == 10
        assert hdul1[0].header['PIXVALUE'] == 1.0

        with pytest.raises(KeyError):
            hdul1[0].header['NAXIS1']

        assert hdul1[0].data is None

        hdul.close()
        hdul1.close()

    def test_StrayPixvalue(self):
        """Regression test for #885
        (https://svn.stsci.edu/trac/ssb/stsci_python/ticket/885)

        Tests that HDUs containing a non-zero NAXIS as well as a PIXVALUE
        keyword in their header are not treated as constant value HDUs.
        """

        data = np.arange(100).reshape((10, 10))
        phdu = fits.PrimaryHDU(data=data)
        hdu = fits.ImageHDU(data=data)

        phdu.header['PIXVALUE'] = 10
        hdu.header['PIXVALUE'] = 10

        hdul = fits.HDUList([phdu, hdu])
        hdul.writeto(self.temp('test.fits'))

        with stpyfits.open(self.temp('test.fits')) as h:
            assert not isinstance(h[0], stpyfits.ConstantValuePrimaryHDU)
            assert not isinstance(h[1], stpyfits.ConstantValueImageHDU)
            assert (h[0].data == data).all()
            assert (h[1].data == data).all()

    def test_DimensionlessConstantValueArray(self):
        """Tests a case that was reported where an HDU can be a constant
        value HDU (it has a PIXVALUE and NAXIS=0) but NPIX1 = NPIX2 = 0 as
        well.
        """

        hdu = stpyfits.PrimaryHDU()
        hdu.header['NAXIS'] = 0
        hdu.header['BITPIX'] = 16
        hdu.header['NPIX1'] = 0
        hdu.header['NPIX2'] = 0
        hdu.header['PIXVALUE'] = 0

        hdu.writeto(self.temp('test.fits'))

        with stpyfits.open(self.temp('test.fits')) as h:
            assert h[0].data is None
            h.writeto(self.temp('test2.fits'))

    def test_DeconvertConstantArray(self):
        """When a constant value array's data is overridden with non-
        constant data, test that when saving the file it removes
        all constant value array keywords and is treated as a normal image
        HDU.
        """

        data = np.ones((100, 100))
        hdu = stpyfits.PrimaryHDU(data=data)
        hdu.header['PIXVALUE'] = 1
        hdu.writeto(self.temp('test.fits'))

        with stpyfits.open(self.temp('test.fits'), mode='update') as h:
            assert h[0].header['PIXVALUE'] == 1
            h[0].data[20:80, 20:80] = 2

        with fits.open(self.temp('test.fits')) as h:
            assert 'PIXVALUE' not in h[0].header
            assert 'NPIX1' not in h[0].header
            assert 'NPIX2' not in h[0].header
            assert h[0].header.count('NAXIS') == 1
            assert h[0].header['NAXIS'] == 2
            assert h[0].header['NAXIS1'] == 100
            assert h[0].header['NAXIS2'] == 100
            assert h[0].data.max() == 2
            assert h[0].data.min() == 1

    def test_GetvalExtensionHDU(self):
        """Regression test for an issue that came up with the fact that
        ImageHDU has a different argument signature from PrimaryHDU.
        """

        data = np.ones((100, 100))
        hdu = stpyfits.ImageHDU(data=data)
        hdu.header['PIXVALUE'] = 1
        hdu.header['FOO'] = 'test'
        hdul = stpyfits.HDUList([stpyfits.PrimaryHDU(), hdu])
        hdul.writeto(self.temp('test.fits'))

        assert stpyfits.getval(self.temp('test.fits'), 'FOO', ext=1) == 'test'



# $Id$

"""
The stpyfits module is an extension to the `astropy.io.fits` module which offers
additional features specific to STScI.  These features include the handling
of Constant Data Value Arrays.

"""
import functools
import numpy as np

import astropy
from astropy.io import fits
# A few imports for backward compatibility; in the earlier stpyfits these were
# overridden, but with fits's new extension system it's not necessary
from astropy.io.fits.util import _is_int
from astropy.utils import lazyproperty
from distutils.version import LooseVersion

ASTROPY_VER_GE32 = LooseVersion(astropy.__version__) >= LooseVersion('3.2')

STPYFITS_ENABLED = False  # Not threadsafe TODO: (should it be?)


# Register the extension classes; simply importing stpyfits does not
# automatically enable it.  Instead, it can be enabled/disabled using these
# functions.
def enable_stpyfits():
    global STPYFITS_ENABLED
    if not STPYFITS_ENABLED:
        fits.register_hdu(ConstantValuePrimaryHDU)
        fits.register_hdu(ConstantValueImageHDU)
        STPYFITS_ENABLED = True


def disable_stpyfits():
    global STPYFITS_ENABLED
    if STPYFITS_ENABLED:
        fits.unregister_hdu(ConstantValuePrimaryHDU)
        fits.unregister_hdu(ConstantValueImageHDU)
        STPYFITS_ENABLED = False


def with_stpyfits(func):
    @functools.wraps(func)
    def wrapped_with_stpyfits(*args, **kwargs):
        global STPYFITS_ENABLED
        was_enabled = STPYFITS_ENABLED
        enable_stpyfits()
        if ASTROPY_VER_GE32:
            from astropy.io.fits.header import _BasicHeader
            fromfile_orig = _BasicHeader.fromfile

            def fromfile_patch(*args):
                raise Exception
            _BasicHeader.fromfile = fromfile_patch

        try:
            # BUG: Forcefully disable lazy loading.
            # Lazy loading breaks ability to initialize ConstantValueHDUs
            # TODO: Investigate the cause upstream (astropy.io.fits)
            if 'write' not in func.__name__:
                kwargs['lazy_load_hdus'] = False
            retval = func(*args, **kwargs)
        finally:
            # Only disable stpyfits if it wasn't already enabled
            if not was_enabled:
                disable_stpyfits()

            if ASTROPY_VER_GE32:
                _BasicHeader.fromfile = fromfile_orig

        return retval
    return wrapped_with_stpyfits


class _ConstantValueImageBaseHDU(fits.hdu.image._ImageBaseHDU):
    """
    A class that extends the `astropy.io.fits.hdu.base._BaseHDU` class to extend its
    behavior to implement STScI specific extensions to `astropy.io.fits`.

    The `astropy.io.fits.hdu.base._BaseHDU class` is:
    """

    __doc__ += fits.hdu.image._ImageBaseHDU.__doc__

    def __init__(self, data=None, header=None, do_not_scale_image_data=False,
                 uint=False, **kwargs):
        if header and 'PIXVALUE' in header and header['NAXIS'] == 0:
            header = header.copy()
            # Add NAXISn keywords for each NPIXn keyword in the header and
            # remove the NPIXn keywords
            naxis = 0
            for card in reversed(header['NPIX*'].cards):
                try:
                    idx = int(card.keyword[len('NPIX'):])
                except ValueError:
                    continue
                hdrlen = len(header)
                header.set('NAXIS' + str(idx), card.value,
                           card.comment, after='NAXIS')
                del header[card.keyword]
                if len(header) < hdrlen:
                    # A blank card was used when updating the header; add the
                    # blank back in.
                    # TODO: Fix header.set so that it has an option not to
                    # use a blank card--this is a detail that we really
                    # shouldn't have to worry about otherwise
                    header.append()

                # Presumably the NPIX keywords are in order of their axis, but
                # just in case somehow they're not...
                naxis = max(naxis, idx)

            # Update the NAXIS keyword with the correct number of axes
            header['NAXIS'] = naxis
        elif header and 'PIXVALUE' in header:
            pixval = header['PIXVALUE']
            if header['BITPIX'] > 0:
                pixval = int(pixval)
            arrayval = self._check_constant_value_data(data)
            if arrayval is not None:
                header = header.copy()
                # Update the PIXVALUE keyword if necessary
                if arrayval != pixval:
                    header['PIXVALUE'] = arrayval
            else:
                header = header.copy()
                # There is a PIXVALUE keyword but NAXIS is not 0 and the data
                # does not match the PIXVALUE.
                # Must remove the PIXVALUE and NPIXn keywords so we recognize
                # that there is non-constant data in the file.
                del header['PIXVALUE']
                for card in header['NPIX*'].cards:
                    try:
                        idx = int(card.keyword[len('NPIX'):])
                    except ValueError:
                        continue
                    del header[card.keyword]

        # Make sure to pass any arguments other than data and header as
        # keyword arguments, because PrimaryHDU and ImageHDU have stupidly
        # different signatures for __init__
        super(_ConstantValueImageBaseHDU, self).__init__(
            data, header, do_not_scale_image_data=do_not_scale_image_data,
            uint=uint)

    @property
    def size(self):
        """
        The HDU's size should always come up as zero so long as there's no
        actual data in it other than the constant value array.
        """

        if 'PIXVALUE' in self._header:
            return 0
        else:
            return super(_ConstantValueImageBaseHDU, self).size

    @lazyproperty
    def data(self):
        if ('PIXVALUE' in self._header and 'NPIX1' not in self._header and
                self._header['NAXIS'] > 0):
            bitpix = self._header['BITPIX']
            dims = self.shape

            # Special case where the pixvalue can be present but all the NPIXn
            # keywords are zero.
            if sum(dims) == 0:
                return None

            code = BITPIX2DTYPE[bitpix]
            pixval = self._header['PIXVALUE']
            if code in ['uint8', 'int16', 'int32', 'int64']:
                pixval = int(pixval)

            raw_data = np.zeros(shape=dims, dtype=code) + pixval

            if raw_data.dtype.str[0] != '>':
                raw_data = raw_data.byteswap(True)

            raw_data.dtype = raw_data.dtype.newbyteorder('>')

            if self._bzero != 0 or self._bscale != 1:
                if bitpix > 16:  # scale integers to Float64
                    data = np.array(raw_data, dtype=np.float64)
                elif bitpix > 0:  # scale integers to Float32
                    data = np.array(raw_data, dtype=np.float32)
                else:  # floating point cases
                    data = raw_data

                if self._bscale != 1:
                    np.multiply(data, self._bscale, data)
                if self._bzero != 0:
                    data += self._bzero

                # delete the keywords BSCALE and BZERO after scaling
                del self._header['BSCALE']
                del self._header['BZERO']
                self._header['BITPIX'] = DTYPE2BITPIX[data.dtype.name]
            else:
                data = raw_data
            return data
        else:
            return super(_ConstantValueImageBaseHDU, self).data

    @data.setter
    def data(self, data):
        self.__dict__['data'] = data
        self._modified = True
        if self.data is not None and not isinstance(data, np.ndarray):
            # Try to coerce the data into a numpy array--this will work, on
            # some level, for most objects
            try:
                data = np.array(data)
            except:
                raise TypeError('data object %r could not be coerced into an '
                                'ndarray' % data)

        if isinstance(data, np.ndarray):
            self._bitpix = DTYPE2BITPIX[data.dtype.name]
            self._axes = list(data.shape)
            self._axes.reverse()
        elif self.data is None:
            self._axes = []
        else:
            raise ValueError('not a valid data array')

        self.update_header()

    @classmethod
    def match_header(cls, header):
        """A constant value HDU will only be recognized as such if the header
        contains a valid PIXVALUE and NAXIS == 0.
        """

        pixvalue = header.get('PIXVALUE')
        naxis = header.get('NAXIS', 0)

        return (super(_ConstantValueImageBaseHDU, cls).match_header(header) and
                (isinstance(pixvalue, float) or _is_int(pixvalue)) and
                naxis == 0)

    def update_header(self):
        if (not self._modified and not self._header._modified and
                (self._has_data and self.shape == self.data.shape)):
            # Not likely that anything needs updating
            return

        super(_ConstantValueImageBaseHDU, self).update_header()

        if 'PIXVALUE' in self._header and self._header['NAXIS'] > 0:
            # This is a Constant Value Data Array.  Verify that the data
            # actually matches the PIXVALUE
            pixval = self._header['PIXVALUE']
            if self._header['BITPIX'] > 0:
                pixval = int(pixval)

            if self.data is None or self.data.nbytes == 0:
                # Empty data array; just keep the existing PIXVALUE
                arrayval = self._header['PIXVALUE']
            else:
                arrayval = self._check_constant_value_data(self.data)
            if arrayval is not None:
                if arrayval != pixval:
                    self._header['PIXVALUE'] = arrayval

                naxis = self._header['NAXIS']
                self._header['NAXIS'] = 0
                for idx in range(naxis, 0, -1):
                    axisval = self._header['NAXIS%d' % idx]
                    self._header.set('NPIX%d' % idx, axisval,
                                     'length of constant array axis %d' % idx,
                                     after='PIXVALUE')
                    del self._header['NAXIS%d' % idx]
            else:
                # No longer a constant value array; remove any remaining
                # NPIX or PIXVALUE keywords
                try:
                    del self._header['PIXVALUE']
                except KeyError:
                    pass

                try:
                    del self._header['NPIX*']
                except KeyError:
                    pass

    def _summary(self):
        summ = super(_ConstantValueImageBaseHDU, self)._summary()
        outsumm = ((summ[0], summ[1],
                    summ[2].replace('ConstantValue', '')) + summ[3:])
        return outsumm

    def _writedata_internal(self, fileobj):
        if 'PIXVALUE' in self._header:
            # This is a Constant Value Data Array, so no data is written
            return 0
        else:
            return super(_ConstantValueImageBaseHDU, self)._writedata_internal(fileobj)

    def _check_constant_value_data(self, data):
        """Verify that the HDU's data is a constant value array."""

        arrayval = data.flat[0]
        if np.all(data == arrayval):
            return arrayval
        return None


class ConstantValuePrimaryHDU(_ConstantValueImageBaseHDU,
                              fits.hdu.PrimaryHDU):
    """Primary HDUs with constant value arrays."""


class ConstantValueImageHDU(_ConstantValueImageBaseHDU, fits.hdu.ImageHDU):
    """Image extension HDUs with constant value arrays."""


# Import the rest of the astropy.io.fits module
from astropy.io.fits import *  # noqa

# For backward-compatibility with older code that thinks PrimaryHDU and
# ImageHDU should support the ConstantValue features
PrimaryHDU = ConstantValuePrimaryHDU
ImageHDU = ConstantValueImageHDU


# Override the other "convenience" functions to use stpyfits
open = fitsopen = with_stpyfits(fits.open)
info = with_stpyfits(fits.info)
append = with_stpyfits(fits.append)
writeto = with_stpyfits(fits.writeto)
update = with_stpyfits(fits.update)
getheader = with_stpyfits(fits.getheader)
getdata = with_stpyfits(fits.getdata)
getval = with_stpyfits(fits.getval)
setval = with_stpyfits(fits.setval)
delval = with_stpyfits(fits.delval)


__all__ = fits.__all__ + ['enable_stpyfits', 'disable_stpyfits',
                          'with_stpyfits', 'ConstantValuePrimaryHDU',
                          'ConstantValueImageHDU']



"""fitsdiff is now a part of Astropy.

Now this module just provides a wrapper around astropy.io.fits.diff for backwards
compatibility with the old interface in case anyone uses it.
"""
import os
import sys

from astropy.io.fits.diff import FITSDiff
from astropy.io.fits.scripts.fitsdiff import log, main


def fitsdiff(input1, input2, comment_excl_list='', value_excl_list='',
             field_excl_list='', maxdiff=10, delta=0.0, neglect_blanks=True,
             output=None):

    if isinstance(comment_excl_list, str):
        comment_excl_list = list_parse(comment_excl_list)

    if isinstance(value_excl_list, str):
        value_excl_list = list_parse(value_excl_list)

    if isinstance(field_excl_list, str):
        field_excl_list = list_parse(field_excl_list)

    diff = FITSDiff(input1, input2, ignore_keywords=value_excl_list,
                    ignore_comments=comment_excl_list,
                    ignore_fields=field_excl_list, numdiffs=maxdiff,
                    tolerance=delta, ignore_blanks=neglect_blanks)

    if output is None:
        output = sys.stdout

    diff.report(output)

    return diff.identical


def list_parse(name_list):
    """Parse a comma-separated list of values, or a filename (starting with @)
    containing a list value on each line.
    """

    if name_list and name_list[0] == '@':
        value = name_list[1:]
        if not os.path.exists(value):
            log.warning('The file %s does not exist' % value)
            return
        try:
            return [v.strip() for v in open(value, 'r').readlines()]
        except IOError as e:
            log.warning('reading %s failed: %s; ignoring this file' %
                        (value, e))
    else:
        return [v.strip() for v in name_list.split(',')]


if __name__ == "__main__":
    sys.exit(main())
